# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
""" undoable transaction objects. """

from cubicweb import RepositoryError
from cubicweb import _

ACTION_LABELS = {
    "C": _("entity creation"),
    "U": _("entity update"),
    "D": _("entity deletion"),
    "A": _("relation add"),
    "R": _("relation removal"),
}


class NoSuchTransaction(RepositoryError):
    # Used by CubicWebException
    msg = _("there is no transaction #%s")

    def __init__(self, txuuid):
        super().__init__(txuuid)
        self.txuuid = txuuid


class Transaction:
    """an undoable transaction"""

    def __init__(self, cnx, uuid, time, ueid):
        self.cnx = cnx
        self.uuid = uuid
        self.datetime = time
        self.user_eid = ueid

    def _execute(self, *args, **kwargs):
        """execute a query using either the req or the cnx"""
        return self.cnx.execute(*args, **kwargs)

    def __repr__(self):
        return f"<Transaction {self.uuid} by {self.user_eid} on {self.datetime}>"

    def user(self):
        """return the user entity which has done the transaction,
        none if not found.
        """
        return self.cnx.find("CWUser", eid=self.user_eid).one()

    def actions_list(self, public=True):
        """return an ordered list of action effectued during that transaction

        if public is true, return only 'public' action, eg not ones triggered
        under the cover by hooks.
        """
        return self.cnx.transaction_actions(self.uuid, public)


class AbstractAction:
    def __init__(self, action, public, order):
        self.action = action
        self.public = public
        self.order = order

    @property
    def label(self):
        return ACTION_LABELS[self.action]

    @property
    def ertype(self):
        """Return the entity or relation type this action is related to"""
        raise NotImplementedError(self)


class EntityAction(AbstractAction):
    def __init__(self, action, public, order, etype, eid, changes):
        super().__init__(action, public, order)
        self.etype = etype
        self.eid = eid
        self.changes = changes

    def __repr__(self):
        return "<{}: {} {} ({})>".format(
            self.label,
            self.eid,
            self.changes,
            self.public and "dbapi" or "hook",
        )

    @property
    def ertype(self):
        """Return the entity or relation type this action is related to"""
        return self.etype


class RelationAction(AbstractAction):
    def __init__(self, action, public, order, rtype, eidfrom, eidto):
        super().__init__(action, public, order)
        self.rtype = rtype
        self.eid_from = eidfrom
        self.eid_to = eidto

    def __repr__(self):
        return "<{}: {} {} {} ({})>".format(
            self.label,
            self.eid_from,
            self.rtype,
            self.eid_to,
            self.public and "dbapi" or "hook",
        )

    @property
    def ertype(self):
        """Return the entity or relation type this action is related to"""
        return self.rtype



# copyright 2003 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""plan execution of rql queries on a single source"""

from typing import Union, TYPE_CHECKING
from rql.nodes import Constant, Relation
from rql.stmts import Delete, Insert, Set, Union as UnionStatement, Select
from yams import KEYWORD_MAP

from cubicweb import QueryError
from cubicweb.rqlrewrite import add_types_restriction, RQLRelationRewriter
from cubicweb.schema import VIRTUAL_RTYPES
from cubicweb.server.edition import EditedEntity

READ_ONLY_RTYPES = {"eid", "has_text", "is", "is_instance_of", "identity"}

_CONSTANT = object()
_FROM_SUBSTEP = object()

if TYPE_CHECKING:
    from cubicweb.server.querier import ExecutionPlan, InsertPlan


def _extract_const_attributes(plan, rqlst, to_build):
    """add constant values to entity def, mark variables to be selected"""
    to_select = {}
    for relation in rqlst.main_relations:
        lhs, rhs = relation.get_variable_parts()
        rtype = relation.r_type
        if rtype in READ_ONLY_RTYPES:
            raise QueryError(f"can't assign to {rtype}")
        try:
            edef = to_build[str(lhs)]
        except KeyError:
            # lhs var is not to build, should be selected and added as an
            # object relation
            edef = to_build[str(rhs)]
            to_select.setdefault(edef, []).append((rtype, lhs, 1))
        else:
            if isinstance(rhs, Constant) and not rhs.uid:
                # add constant values to entity def
                eschema = edef.entity.e_schema
                attrtype = eschema.subject_relations[rtype].objects(eschema)[0]
                value = _eval_const_attribute(rhs, plan.args, attrtype)
                if attrtype == "Password" and isinstance(value, str):
                    value = value.encode("UTF8")
                edef.edited_attribute(rtype, value)
            elif str(rhs) in to_build:
                # create a relation between two newly created variables
                plan.add_relation_def((edef, rtype, to_build[rhs.name]))
            else:
                to_select.setdefault(edef, []).append((rtype, rhs, 0))
    return to_select


def _extract_eid_consts(plan, rqlst):
    """return a dict mapping rqlst variable object to their eid if specified in
    the syntax tree
    """
    cnx = plan.cnx
    if rqlst.where is None:
        return {}
    eidconsts = {}
    neweids = cnx.transaction_data.get("neweids", ())
    checkread = cnx.read_security
    for rel in rqlst.where.get_nodes(Relation):
        # only care for 'eid' relations ...
        if (
            rel.r_type == "eid"
            # ... that are not part of a NOT clause ...
            and not rel.neged(strict=True)
            # ... and where eid is specified by '=' operator.
            and rel.children[1].operator == "="
        ):
            lhs, rhs = rel.get_variable_parts()
            if isinstance(rhs, Constant):
                eid = int(rhs.eval(plan.args))
                # check read permission here since it may not be done by
                # the generated select substep if not emited (eg nothing
                # to be selected)
                if checkread and eid not in neweids:
                    with cnx.security_enabled(read=False):
                        cnx.vreg.schema.entity_schema_for(
                            cnx.entity_type(eid)
                        ).check_perm(cnx, "read", eid=eid)
                eidconsts[lhs.variable] = eid
    return eidconsts


def _build_substep_query(select, origrqlst):
    """Finalize substep select query that should be executed to get proper
    selection of stuff to insert/update.

    Return None when no query actually needed, else the given select node that
    will be used as substep query.
    """
    if origrqlst.where is not None and not select.selection:
        # no selection, append one randomly by searching for a relation which is
        # not neged neither a type restriction (is/is_instance_of)
        for rel in origrqlst.where.iget_nodes(Relation):
            if not (rel.neged(traverse_scope=True) or rel.is_types_restriction()):
                select.append_selected(rel.children[0].copy(select))
                break
        else:
            return None
    if select.selection:
        if origrqlst.where is not None:
            select.set_where(origrqlst.where.copy(select))
        if getattr(origrqlst, "having", None):
            select.set_having([sq.copy(select) for sq in origrqlst.having])
        return select
    return None


class SSPlanner:
    """SingleSourcePlanner: build execution plan for rql queries

    optimized for single source repositories
    """

    def __init__(self, schema, rqlhelper):
        self.schema = schema
        self.rqlhelper = rqlhelper

    def build_plan(self, plan: Union["ExecutionPlan", "InsertPlan"]) -> None:
        """build an execution plan from a RQL query

        do nothing here, dispatch according to the statement type
        """
        if isinstance(plan.rqlst, Set):
            steps = self.build_set_plan(plan, plan.rqlst)
        elif isinstance(plan.rqlst, Delete):
            steps = self.build_delete_plan(plan, plan.rqlst)
        elif isinstance(plan.rqlst, UnionStatement):
            steps = self.build_select_plan(plan, plan.rqlst)
        elif isinstance(plan.rqlst, Insert):
            steps = self.build_insert_plan(plan, plan.rqlst)
        else:
            raise ValueError("plan is not associated to a valid statement")

        for step in steps:
            plan.add_step(step)

    def build_select_plan(self, plan, rqlst):
        """build execution plan for a SELECT RQL query. Suppose only one source
        is available and so avoid work need for query decomposition among sources

        the rqlst should not be tagged at this point.
        """
        plan.preprocess(rqlst)
        return (OneFetchStep(plan, rqlst),)

    def build_insert_plan(self, plan, rqlst):
        """get an execution plan from an INSERT RQL query"""
        # each variable in main variables is a new entity to insert
        to_build = {}
        cnx = plan.cnx
        etype_class = cnx.vreg["etypes"].etype_class
        for etype, var in rqlst.main_variables:
            # need to do this since entity class is shared w. web client code !
            to_build[var.name] = EditedEntity(etype_class(etype)(cnx))
            plan.add_entity_def(to_build[var.name])
        # add constant values to entity def, mark variables to be selected
        to_select = _extract_const_attributes(plan, rqlst, to_build)
        # add necessary steps to add relations and update attributes
        step = InsertStep(plan)  # insert each entity and its relations
        step.children += self._compute_relation_steps(plan, rqlst, to_select)
        return (step,)

    def _compute_relation_steps(self, plan, rqlst, to_select):
        """handle the selection of relations for an insert query"""
        eidconsts = _extract_eid_consts(plan, rqlst)
        for edef, rdefs in to_select.items():
            # create a select rql st to fetch needed data
            select = Select()
            eschema = edef.entity.e_schema
            for i, (rtype, term, reverse) in enumerate(rdefs):
                if getattr(term, "variable", None) in eidconsts:
                    value = eidconsts[term.variable]
                else:
                    select.append_selected(term.copy(select))
                    value = _FROM_SUBSTEP
                if reverse:
                    rdefs[i] = (rtype, InsertRelationsStep.REVERSE_RELATION, value)
                else:
                    rschema = eschema.subject_relations[rtype]
                    if rschema.final or rschema.inlined:
                        rdefs[i] = (rtype, InsertRelationsStep.FINAL, value)
                    else:
                        rdefs[i] = (rtype, InsertRelationsStep.RELATION, value)
            step = InsertRelationsStep(plan, edef, rdefs)
            select = _build_substep_query(select, rqlst)
            if select is not None:
                step.children += self._select_plan(plan, select, rqlst.solutions)
            yield step

    def build_delete_plan(self, plan, rqlst):
        """get an execution plan from a DELETE RQL query"""
        # build a select query to fetch entities to delete
        steps = []
        for etype, var in rqlst.main_variables:
            step = DeleteEntitiesStep(plan)
            step.children += self._sel_variable_step(plan, rqlst, etype, var)
            steps.append(step)
        for relation in rqlst.main_relations:
            rtype = relation.r_type
            if self.schema[rtype].rule:
                raise QueryError(f"'{rtype}' is a computed relation")
            step = DeleteRelationsStep(plan, rtype)
            step.children += self._sel_relation_steps(plan, rqlst, relation)
            steps.append(step)
        return steps

    def _sel_variable_step(self, plan, rqlst, etype, varref):
        """handle the selection of variables for a delete query"""
        select = Select()
        varref = varref.copy(select)
        select.defined_vars = {varref.name: varref.variable}
        select.append_selected(varref)
        if rqlst.where is not None:
            select.set_where(rqlst.where.copy(select))
        if getattr(rqlst, "having", None):
            select.set_having([x.copy(select) for x in rqlst.having])
        if etype != "Any":
            select.add_type_restriction(varref.variable, etype)
        return self._select_plan(plan, select, rqlst.solutions)

    def _sel_relation_steps(self, plan, rqlst, relation):
        """handle the selection of relations for a delete query"""
        select = Select()
        lhs, rhs = relation.get_variable_parts()
        select.append_selected(lhs.copy(select))
        select.append_selected(rhs.copy(select))
        select.set_where(relation.copy(select))
        if rqlst.where is not None:
            select.add_restriction(rqlst.where.copy(select))
        if getattr(rqlst, "having", None):
            select.set_having([x.copy(select) for x in rqlst.having])
        return self._select_plan(plan, select, rqlst.solutions)

    def build_set_plan(self, plan, rqlst):
        """get an execution plan from an SET RQL query"""
        getrschema = self.schema.relation_schema_for
        select = Select()  # potential substep query
        selectedidx = {}  # local state
        updatedefs = []  # definition of update attributes/relations
        selidx = residx = 0  # substep selection / resulting rset indexes
        # search for eid const in the WHERE clause
        eidconsts = _extract_eid_consts(plan, rqlst)
        # build `updatedefs` describing things to update and add necessary
        # variables to the substep selection
        for i, relation in enumerate(rqlst.main_relations):
            if relation.r_type in VIRTUAL_RTYPES:
                raise QueryError(f"can not assign to {relation.r_type!r} relation")
            lhs, rhs = relation.get_variable_parts()
            lhskey = lhs.as_string()
            if lhskey not in selectedidx:
                if lhs.variable in eidconsts:
                    eid = eidconsts[lhs.variable]
                    lhsinfo = (_CONSTANT, eid, residx)
                else:
                    select.append_selected(lhs.copy(select))
                    lhsinfo = (_FROM_SUBSTEP, selidx, residx)
                    selidx += 1
                residx += 1
                selectedidx[lhskey] = lhsinfo
            else:
                lhsinfo = selectedidx[lhskey][:-1] + (None,)
            rschema = getrschema(relation.r_type)
            rhskey = rhs.as_string()
            if rhskey not in selectedidx:
                if isinstance(rhs, Constant):
                    attrtype = rschema.objects()[0]
                    rhsinfo = (
                        _CONSTANT,
                        _eval_const_attribute(rhs, plan.args, attrtype),
                        residx,
                    )
                elif getattr(rhs, "variable", None) in eidconsts:
                    eid = eidconsts[rhs.variable]
                    rhsinfo = (_CONSTANT, eid, residx)
                else:
                    select.append_selected(rhs.copy(select))
                    rhsinfo = (_FROM_SUBSTEP, selidx, residx)
                    selidx += 1
                residx += 1
                selectedidx[rhskey] = rhsinfo
            else:
                rhsinfo = selectedidx[rhskey][:-1] + (None,)
            updatedefs.append((lhsinfo, rhsinfo, rschema))
        # the update step
        step = UpdateStep(plan, updatedefs)
        # when necessary add substep to fetch yet unknown values
        select = _build_substep_query(select, rqlst)
        if select is not None:
            # set distinct to avoid potential duplicate key error
            select.distinct = True
            step.children += self._select_plan(plan, select, rqlst.solutions)
        return (step,)

    # internal methods ########################################################

    def _select_plan(self, plan, select, solutions):
        union = UnionStatement()
        union.append(select)
        select.clean_solutions(solutions)
        add_types_restriction(self.schema, select)
        # Rewrite computed relations
        rewriter = RQLRelationRewriter(plan.cnx)
        rewriter.rewrite(union, plan.args)
        self.rqlhelper.annotate(union)
        return self.build_select_plan(plan, union)


def prepare_plan(plan, schema, rqlhelper):
    """Add steps to a plan to prepare it for execution"""
    return SSPlanner(schema, rqlhelper).build_plan(plan)


# execution steps and helper functions ########################################


class Step:
    """base abstract class for execution step"""

    def __init__(self, plan):
        self.plan = plan
        self.children = []

    def execute_child(self):
        assert len(self.children) == 1
        return self.children[0].execute()

    def execute_children(self):
        for step in self.children:
            step.execute()

    def execute(self):
        """execute this step and store partial (eg this step) results"""
        raise NotImplementedError()

    def mytest_repr(self):
        """return a representation of this step suitable for test"""
        return (self.__class__.__name__,)

    def test_repr(self):
        """return a representation of this step suitable for test"""
        return self.mytest_repr() + ([step.test_repr() for step in self.children],)


class OneFetchStep(Step):
    """step consisting in fetching data from sources and directly returning
    results
    """

    def __init__(self, plan, union):
        Step.__init__(self, plan)
        self.union = union
        self.rql_query_tracing_token = None

    def execute(self):
        """call .syntax_tree_search with the given syntax tree on each
        source for each solution
        """
        self.execute_children()
        cnx = self.plan.cnx
        args = self.plan.args
        union = self.union
        if self.plan.cache_key is None:
            cachekey = None
        # union may have been splited into subqueries, in which case we can't
        # use plan.cache_key, rebuild a cache key
        elif isinstance(self.plan.cache_key, tuple):
            cachekey = list(self.plan.cache_key)
            cachekey[0] = union.as_string()
            cachekey = tuple(cachekey)
        else:
            cachekey = union.as_string()
        # get results for query
        source = cnx.repo.system_source
        result = source.syntax_tree_search(
            cnx,
            union,
            args,
            cachekey,
            rql_query_tracing_token=self.rql_query_tracing_token,
        )
        return result

    def mytest_repr(self):
        """return a representation of this step suitable for test"""
        return (
            self.__class__.__name__,
            sorted(
                (r.as_string(kwargs=self.plan.args), r.solutions)
                for r in self.union.children
            ),
        )


# UPDATE/INSERT/DELETE steps ##################################################


class InsertRelationsStep(Step):
    """step consisting in adding attributes/relations to entity defs from a
    previous FetchStep

    relations values comes from the latest result, with one columns for
    each relation defined in self.rdefs

    for one entity definition, we'll construct N entity, where N is the
    number of the latest result
    """

    FINAL = 0
    RELATION = 1
    REVERSE_RELATION = 2

    def __init__(self, plan, edef, rdefs):
        Step.__init__(self, plan)
        # partial entity definition to expand
        self.edef = edef
        # definition of relations to complete
        self.rdefs = rdefs

    def execute(self):
        """execute this step"""
        base_edef = self.edef
        edefs = []
        if self.children:
            result = self.execute_child()
        else:
            result = [[]]
        for row in result:
            # get a new entity definition for this row
            edef = base_edef.clone()
            # complete this entity def using row values
            index = 0
            for rtype, rorder, value in self.rdefs:
                if value is _FROM_SUBSTEP:
                    value = row[index]
                    index += 1
                if rorder == InsertRelationsStep.FINAL:
                    edef.edited_attribute(rtype, value)
                elif rorder == InsertRelationsStep.RELATION:
                    self.plan.add_relation_def((edef, rtype, value))
                    edef.querier_pending_relations[(rtype, "subject")] = value
                else:
                    self.plan.add_relation_def((value, rtype, edef))
                    edef.querier_pending_relations[(rtype, "object")] = value
            edefs.append(edef)
        self.plan.substitute_entity_def(base_edef, edefs)
        return result


class InsertStep(Step):
    """step consisting in inserting new entities / relations"""

    def execute(self):
        """execute this step"""
        for step in self.children:
            assert isinstance(step, InsertRelationsStep)
            step.plan = self.plan
            step.execute()
        # insert entities first
        result = self.plan.insert_entity_defs()
        # then relation
        self.plan.insert_relation_defs()
        # return eids of inserted entities
        return result


class DeleteEntitiesStep(Step):
    """step consisting in deleting entities"""

    def execute(self):
        """execute this step"""
        results = self.execute_child()
        if results:
            todelete = frozenset(int(eid) for eid, in results)
            cnx = self.plan.cnx
            cnx.repo.glob_delete_entities(cnx, todelete)
        return results


class DeleteRelationsStep(Step):
    """step consisting in deleting relations"""

    def __init__(self, plan, rtype):
        Step.__init__(self, plan)
        self.rtype = rtype

    def execute(self):
        """execute this step"""
        cnx = self.plan.cnx
        delete = cnx.repo.glob_delete_relation
        for subj, obj in self.execute_child():
            delete(cnx, subj, self.rtype, obj)


class UpdateStep(Step):
    """step consisting in updating entities / adding relations from relations
    definitions and from results fetched in previous step
    """

    def __init__(self, plan, updatedefs):
        Step.__init__(self, plan)
        self.updatedefs = updatedefs

    def execute(self):
        """execute this step"""
        cnx = self.plan.cnx
        repo = cnx.repo
        edefs = {}
        relations = {}
        # insert relations
        if self.children:
            result = self.execute_child()
        else:
            result = [[]]
        for i, row in enumerate(result):
            newrow = []
            for lhsinfo, rhsinfo, rschema in self.updatedefs:
                if rschema.rule:
                    raise QueryError(f"'{rschema.type}' is a computed relation")
                lhsval = _handle_relterm(lhsinfo, row, newrow)
                rhsval = _handle_relterm(rhsinfo, row, newrow)
                if rschema.final or rschema.inlined:
                    eid = int(lhsval)
                    try:
                        edited = edefs[eid]
                    except KeyError:
                        edef = cnx.entity_from_eid(eid)
                        edefs[eid] = edited = EditedEntity(edef)
                    edited.edited_attribute(str(rschema), rhsval)
                else:
                    str_rschema = str(rschema)
                    if str_rschema in relations:
                        relations[str_rschema].append((lhsval, rhsval))
                    else:
                        relations[str_rschema] = [(lhsval, rhsval)]
            result[i] = newrow
        # update entities
        repo.glob_add_relations(cnx, relations)
        for eid, edited in edefs.items():
            repo.glob_update_entity(cnx, edited)
        return result


def _handle_relterm(info, row, newrow):
    if info[0] is _CONSTANT:
        val = info[1]
    else:  # _FROM_SUBSTEP
        val = row[info[1]]
    if info[-1] is not None:
        newrow.append(val)
    return val


def _eval_const_attribute(const_node, kwargs, attrtype):
    if const_node.type == "Substitute":
        return kwargs[const_node.value]
    if const_node.type in ["Date", "Datetime"]:
        assert attrtype in ["Datetime", "TZDatetime", "Date"]
        return KEYWORD_MAP[attrtype][const_node.value]()
    else:
        return const_node.value



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""user interface libraries

contains some functions designed to help implementation of cubicweb user
interface.
"""


import csv
import re
import traceback
from io import StringIO
from datetime import timedelta
from functools import wraps

from logilab.common.date import ustrftime
from logilab.mtconverter import xml_escape, html_unescape
from lxml import etree
from lxml.html import defs, clean

from cubicweb import _
from cubicweb.utils import js_dumps
from cubicweb.ext.markdown import markdown_publish  # noqa


def rql_for_eid(eid):
    """return the rql query necessary to fetch entity with the given eid.  This
    function should only be used to generate link with rql inside, not to give
    to cursor.execute (in which case you won't benefit from rql cache).

    :Parameters:
      - `eid`: the eid of the entity we should search
    :rtype: str
    :return: the rql query
    """
    return f"Any X WHERE X eid {eid}"


def eid_param(name, eid):
    assert name is not None
    assert eid is not None
    return f"{name}:{eid}"


def print_bytes(value, req, props, displaytime=True):
    return ""


def print_string(value, req, props, displaytime=True):
    # don't translate empty value if you don't want strange results
    if props is not None and value and props.get("internationalizable"):
        return req._(value)
    return value


def print_int(value, req, props, displaytime=True):
    return str(value)


def print_date(value, req, props, displaytime=True):
    return ustrftime(value, req.property_value("ui.date-format"))


def print_time(value, req, props, displaytime=True):
    return ustrftime(value, req.property_value("ui.time-format"))


def print_tztime(value, req, props, displaytime=True):
    return ustrftime(value, req.property_value("ui.time-format")) + " UTC"


def print_datetime(value, req, props, displaytime=True):
    if displaytime:
        return ustrftime(value, req.property_value("ui.datetime-format"))
    return ustrftime(value, req.property_value("ui.date-format"))


def print_tzdatetime(value, req, props, displaytime=True):
    if displaytime:
        return ustrftime(value, req.property_value("ui.datetime-format")) + " UTC"
    return ustrftime(value, req.property_value("ui.date-format"))


_("%d years")
_("%d months")
_("%d weeks")
_("%d days")
_("%d hours")
_("%d minutes")
_("%d seconds")


def print_timedelta(value, req, props, displaytime=True):
    if isinstance(value, int):
        # `date - date`, unlike `datetime - datetime` gives an int
        # (number of days), not a timedelta
        # XXX should rql be fixed to return Int instead of Interval in
        #     that case? that would be probably the proper fix but we
        #     loose information on the way...
        value = timedelta(days=value)
    if value.days > 730 or value.days < -730:  # 2 years
        return req._("%d years") % (value.days // 365)
    elif value.days > 60 or value.days < -60:  # 2 months
        return req._("%d months") % (value.days // 30)
    elif value.days > 14 or value.days < -14:  # 2 weeks
        return req._("%d weeks") % (value.days // 7)
    elif value.days > 2 or value.days < -2:
        return req._("%d days") % int(value.days)
    else:
        minus = 1 if value.days >= 0 else -1
        if value.seconds > 3600:
            return req._("%d hours") % (int(value.seconds // 3600) * minus)
        elif value.seconds >= 120:
            return req._("%d minutes") % (int(value.seconds // 60) * minus)
        else:
            return req._("%d seconds") % (int(value.seconds) * minus)


def print_boolean(value, req, props, displaytime=True):
    if value:
        return req._("yes")
    return req._("no")


def print_float(value, req, props, displaytime=True):
    return str(req.property_value("ui.float-format") % value)  # XXX cast needed ?


PRINTERS = {
    "Bytes": print_bytes,
    "String": print_string,
    "Int": print_int,
    "BigInt": print_int,
    "Date": print_date,
    "Time": print_time,
    "TZTime": print_tztime,
    "Datetime": print_datetime,
    "TZDatetime": print_tzdatetime,
    "Boolean": print_boolean,
    "Float": print_float,
    "Decimal": print_float,
    "Interval": print_timedelta,
}


def css_em_num_value(vreg, propname, default):
    """we try to read an 'em' css property
    if we get another unit we're out of luck and resort to the given default
    (hence, it is strongly advised not to specify but ems for this css prop)
    """
    propvalue = vreg.config.uiprops[propname].lower().strip()
    if propvalue.endswith("em"):
        try:
            return float(propvalue[:-2])
        except Exception:
            vreg.warning("css property %s looks malformed (%r)", propname, propvalue)
    else:
        vreg.warning(
            "css property %s should use em (currently is %r)", propname, propvalue
        )
    return default


# text publishing #############################################################


try:
    from cubicweb_web.ext.rest import rest_publish  # pylint: disable=W0611
except ImportError:

    def rest_publish(entity, data):
        """default behaviour if docutils was not found"""
        return xml_escape(data)


TAG_PROG = re.compile(r"</?.*?>", re.U)


def remove_html_tags(text):
    """Removes HTML tags from text

    >>> remove_html_tags('<td>hi <a href="http://www.google.fr">world</a></td>')
    'hi world'
    >>>
    """
    return TAG_PROG.sub("", text)


REF_PROG = re.compile(r"<ref\s+rql=([\'\"])([^\1]*?)\1\s*>([^<]*)</ref>", re.U)


def _subst_rql(view, obj):
    delim, rql, descr = obj.groups()
    return f'<a href="{view._cw.build_url(rql=rql)}">{descr}</a>'


def html_publish(view, text):
    """replace <ref rql=''> links by <a href="...">"""
    if not text:
        return ""
    return REF_PROG.sub(lambda obj, view=view: _subst_rql(view, obj), text)


# fallback implementation, nicer one defined below if lxml> 2.0 is available
def safe_cut(text, length):
    """returns a string of length <length> based on <text>, removing any html
    tags from given text if cut is necessary."""
    if text is None:
        return ""
    noenttext = html_unescape(text)
    text_nohtml = remove_html_tags(noenttext)
    # try to keep html tags if text is short enough
    if len(text_nohtml) <= length:
        return text
    # else if un-tagged text is too long, cut it
    return xml_escape(text_nohtml[:length] + "...")


fallback_safe_cut = safe_cut

REM_ROOT_HTML_TAGS = re.compile("</(body|html)>", re.U)


ALLOWED_TAGS = (
    defs.general_block_tags
    | defs.list_tags
    | defs.table_tags
    | defs.phrase_tags
    | defs.font_style_tags
    | {"span", "a", "br", "img", "map", "area", "sub", "sup", "canvas"}
)

CLEANER = clean.Cleaner(
    allow_tags=ALLOWED_TAGS,
    remove_unknown_tags=False,
    style=True,
    safe_attrs_only=True,
    add_nofollow=False,
)


def soup2xhtml(data, encoding):
    """tidy html soup by allowing some element tags and return the result"""
    # remove spurious </body> and </html> tags, then normalize line break
    # (see http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7.1)
    data = REM_ROOT_HTML_TAGS.sub("", "\n".join(data.splitlines()))
    xmltree = etree.HTML(CLEANER.clean_html(f"<div>{data}</div>"))
    # NOTE: lxml 2.0 does support encoding='unicode', but last time I (syt)
    # tried I got weird results (lxml 2.2.8)
    body = etree.tostring(xmltree[0], encoding=encoding)
    # remove <body> and </body> and decode to unicode
    snippet = body[6:-7].decode(encoding)
    # take care to bad xhtml (for instance starting with </div>) which
    # may mess with the <div> we added below. Only remove it if it's
    # still there...
    if snippet.startswith("<div>") and snippet.endswith("</div>"):
        snippet = snippet[5:-6]
    return snippet

    # lxml.Cleaner envelops text elements by internal logic (not accessible)
    # see http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7.1
    # TODO drop attributes in elements
    # TODO add policy configuration (content only, embedded content, ...)
    # XXX this is buggy for "<p>text1</p><p>text2</p>"...
    # XXX drop these two snippets action and follow the lxml behaviour
    # XXX (tests need to be updated)
    # if snippet.startswith('<div>') and snippet.endswith('</div>'):
    #     snippet = snippet[5:-6]
    # if snippet.startswith('<p>') and snippet.endswith('</p>'):
    #     snippet = snippet[3:-4]
    return snippet.decode(encoding)


if hasattr(etree.HTML("<div>test</div>"), "iter"):  # XXX still necessary?
    # pylint: disable=E0102
    def safe_cut(text, length):
        """returns an html document of length <length> based on <text>,
        and cut is necessary.
        """
        if text is None:
            return ""
        dom = etree.HTML(text)
        curlength = 0
        add_ellipsis = False
        for element in dom.iter():
            if curlength >= length:
                parent = element.getparent()
                parent.remove(element)
                if curlength == length and (element.text or element.tail):
                    add_ellipsis = True
            else:
                if element.text is not None:
                    element.text = cut(element.text, length - curlength)
                    curlength += len(element.text)
                if element.tail is not None:
                    if curlength < length:
                        element.tail = cut(element.tail, length - curlength)
                        curlength += len(element.tail)
                    elif curlength == length:
                        element.tail = "..."
                    else:
                        element.tail = ""
        text = etree.tounicode(dom[0])[6:-7]  # remove wrapping <body></body>
        if add_ellipsis:
            return text + "..."
        return text


def text_cut(text, nbwords=30, gotoperiod=True):
    """from the given plain text, return a text with at least <nbwords> words,
    trying to go to the end of the current sentence.

    :param nbwords: the minimum number of words required
    :param gotoperiod: specifies if the function should try to go to
                       the first period after the cut (i.e. finish
                       the sentence if possible)

    Note that spaces are normalized.
    """
    if text is None:
        return ""
    words = text.split()
    text = " ".join(words)  # normalize spaces
    textlength = minlength = len(" ".join(words[:nbwords]))
    if gotoperiod:
        textlength = text.find(".", minlength) + 1
        if textlength == 0:  # no period found
            textlength = minlength
    return text[:textlength]


def cut(text, length):
    """returns a string of a maximum length <length> based on <text>
    (approximatively, since if text has been  cut, '...' is added to the end of the string,
    resulting in a string of len <length> + 3)
    """
    if text is None:
        return ""
    if len(text) <= length:
        return text
    # else if un-tagged text is too long, cut it
    return text[:length] + "..."


# HTML generation helper functions ############################################


class _JSId:
    def __init__(self, id, parent=None):
        self.id = id
        self.parent = parent

    def __str__(self):
        if self.parent:
            return f"{self.parent}.{self.id}"
        return str(self.id)

    def __getattr__(self, attr):
        return _JSId(attr, self)

    def __call__(self, *args):
        return _JSCallArgs(args, self)


class _JSCallArgs(_JSId):
    def __init__(self, args, parent=None):
        assert isinstance(args, tuple)
        self.args = args
        self.parent = parent

    def __str__(self):
        args = []
        for arg in self.args:
            args.append(js_dumps(arg))
        if self.parent:
            return f"{self.parent}({','.join(args)})"
        return ",".join(args)


class _JS:
    def __getattr__(self, attr):
        return _JSId(attr)


js = _JS()
js.__doc__ = """\
magic object to return strings suitable to call some javascript function with
the given arguments (which should be correctly typed).

>>> str(js.pouet(1, "2"))
'pouet(1,"2")'
>>> str(js.cw.pouet(1, "2"))
'cw.pouet(1,"2")'
>>> str(js.cw.pouet(1, "2").pouet(None))
'cw.pouet(1,"2").pouet(null)'
>>> str(js.cw.pouet(1, JSString("$")).pouet(None))
'cw.pouet(1,$).pouet(null)'
>>> str(js.cw.pouet(1, {'callback': JSString("cw.cb")}).pouet(None))
'cw.pouet(1,{callback: cw.cb}).pouet(null)'
"""


def domid(string):
    """return a valid DOM id from a string (should also be usable in jQuery
    search expression...)
    """
    return string.replace(".", "_").replace("-", "_")


HTML4_EMPTY_TAGS = frozenset(
    ("base", "meta", "link", "hr", "br", "param", "img", "area", "input", "col")
)


def sgml_attributes(attrs):
    return " ".join(
        f'{attr}="{xml_escape(str(value))}"'
        for attr, value in sorted(attrs.items())
        if value is not None
    )


def simple_sgml_tag(tag, content=None, escapecontent=True, **attrs):
    """generation of a simple sgml tag (eg without children tags) easier

    content and attri butes will be escaped
    """
    value = f"<{tag}"
    if attrs:
        try:
            attrs["class"] = attrs.pop("klass")
        except KeyError:
            pass
        value += " " + sgml_attributes(attrs)
    if content:
        if escapecontent:
            content = xml_escape(str(content))
        value += f">{content}</{tag}>"
    else:
        if tag in HTML4_EMPTY_TAGS:
            value += " />"
        else:
            value += f"></{tag}>"
    return value


def tooltipize(text, tooltip, url=None):
    """make an HTML tooltip"""
    url = url or "#"
    return f'<a href="{url}" title="{tooltip}">{text}</a>'


def toggle_action(nodeid):
    """builds a HTML link that uses the js toggleVisibility function"""
    return f"javascript: toggleVisibility('{nodeid}')"


def toggle_link(nodeid, label):
    """builds a HTML link that uses the js toggleVisibility function"""
    return f'<a href="{toggle_action(nodeid)}">{label}</a>'


def ureport_as_html(layout):
    from logilab.common.ureports import HTMLWriter

    formater = HTMLWriter(True)
    stream = StringIO()  # UStringIO() don't want unicode assertion
    formater.format(layout, stream)
    res = stream.getvalue()
    if isinstance(res, bytes):
        res = res.decode("UTF8")
    return res


# traceback formatting ########################################################


def exc_message(ex, encoding):
    excmsg = str(ex)
    exctype = ex.__class__.__name__
    return f"{exctype}: {excmsg}"


def rest_traceback(info, exception):
    """return a unicode ReST formated traceback"""
    res = ["Traceback\n---------\n::\n"]
    for stackentry in traceback.extract_tb(info[2]):
        res.append("\tFile %s, line %s, function %s" % tuple(stackentry[:3]))
        if stackentry[3]:
            data = xml_escape(stackentry[3])
            res.append(f"\t  {data}")
    res.append("\n")
    try:
        res.append(f"\t Error: {exception}\n")
    except Exception:
        pass
    return "\n".join(res)


def html_traceback(info, exception, title="", encoding="ISO-8859-1", body=""):
    """return an html formatted traceback from python exception infos."""
    tcbk = info[2]
    stacktb = traceback.extract_tb(tcbk)
    strings = []
    if body:
        strings.append('<div class="error_body">')
        # FIXME
        strings.append(body)
        strings.append("</div>")
    if title:
        strings.append(f'<h1 class="error">{xml_escape(title)}</h1>')
    try:
        strings.append(
            '<p class="error">%s</p>'
            % xml_escape(str(exception)).replace("\n", "<br />")
        )
    except UnicodeError:
        pass
    strings.append('<div class="error_traceback">')
    for index, stackentry in enumerate(stacktb):
        strings.append(
            '<b>File</b> <b class="file">%s</b>, <b>line</b> '
            '<b class="line">%s</b>, <b>function</b> '
            '<b class="function">%s</b>:<br/>'
            % (xml_escape(stackentry[0]), stackentry[1], xml_escape(stackentry[2]))
        )
        if stackentry[3]:
            string = xml_escape(stackentry[3])
            strings.append(f"&#160;&#160;{string}<br/>\n")
        # add locals info for each entry
        try:
            local_context = tcbk.tb_frame.f_locals
            html_info = []
            chars = 0
            for name, value in local_context.items():
                value = xml_escape(repr(value))
                info = f'<span class="name">{name}</span>={value}, '
                line_length = len(name) + len(value)
                chars += line_length
                # 150 is the result of *years* of research ;-) (CSS might be helpful here)
                if chars > 150:
                    info = "<br/>" + info
                    chars = line_length
                html_info.append(info)
            boxid = "ctxlevel%d" % index
            strings.append(f"[{toggle_link(boxid, '+')}]")
            strings.append(
                '<div id="%s" class="pycontext hidden">%s</div>'
                % (boxid, "".join(html_info))
            )
            tcbk = tcbk.tb_next
        except Exception:
            pass  # doesn't really matter if we have no context info
    strings.append("</div>")
    return "\n".join(strings)


# csv files / unicode support #################################################


class UnicodeCSVWriter:
    """proxies calls to csv.writer.writerow to be able to deal with unicode

    Under Python 3, this code no longer encodes anything."""

    def __init__(self, wfunc, encoding, **kwargs):
        self.writer = csv.writer(self, **kwargs)
        self.wfunc = wfunc
        self.encoding = encoding

    def write(self, data):
        self.wfunc(data)

    def writerow(self, row):
        self.writer.writerow(row)
        return

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


# some decorators #############################################################


class limitsize:
    def __init__(self, maxsize):
        self.maxsize = maxsize

    def __call__(self, function):
        @wraps(function)
        def newfunc(*args, **kwargs):
            ret = function(*args, **kwargs)
            if isinstance(ret, str):
                return ret[: self.maxsize]
            return ret

        return newfunc


def htmlescape(function):
    @wraps(function)
    def newfunc(*args, **kwargs):
        ret = function(*args, **kwargs)
        assert isinstance(ret, str)
        return xml_escape(ret)

    return newfunc



from yams.buildobjs import EntityType, SubjectRelation

from cubicweb.schemas.base import in_group, RELATION_MANAGERS_PERMISSIONS

# copy __permissions__ to avoid modifying a shared dictionary
in_group.__permissions__ = in_group.__permissions__.copy()
in_group.__permissions__["read"] = ("managers",)


class CWSourceSchemaConfig(EntityType):
    cw_for_source = SubjectRelation(
        "CWSource",
        inlined=True,
        cardinality="1*",
        composite="object",
        __permissions__=RELATION_MANAGERS_PERMISSIONS,
    )


cw_for_source = CWSourceSchemaConfig.get_relation("cw_for_source")
cw_for_source.__permissions__ = {
    "read": ("managers", "users"),
    "add": ("managers",),
    "delete": ("managers",),
}



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Tests for notification sobjects"""
from cubicweb.devtools import BASE_URL
from cubicweb.devtools.testlib import CubicWebTC, MAILBOX


class NotificationTC(CubicWebTC):
    def test_recipients_finder(self):
        with self.admin_access.cnx() as cnx:
            urset = cnx.execute('CWUser X WHERE X login "admin"')
            cnx.execute(
                'INSERT EmailAddress X: X address "admin@logilab.fr", U primary_email X'
                " WHERE U eid %(x)s",
                {"x": urset[0][0]},
            )
            cnx.execute(
                'INSERT CWProperty X: X pkey "ui.language", X value "fr", X for_user U '
                "WHERE U eid %(x)s",
                {"x": urset[0][0]},
            )
            cnx.commit()  # commit so that admin get its properties updated
            finder = self.vreg["services"].select("recipients_finder", cnx, rset=urset)
            self.set_option("default-recipients-mode", "none")
            self.assertEqual(finder.recipients(), [])
            self.set_option("default-recipients-mode", "users")
            self.assertEqual(finder.recipients(), [cnx.user])
            self.set_option("default-recipients-mode", "default-dest-addrs")
            self.set_option("default-dest-addrs", "abcd@logilab.fr, efgh@logilab.fr")
            self.assertEqual(
                list(finder.recipients()),
                [("abcd@logilab.fr", "en"), ("efgh@logilab.fr", "en")],
            )

    def test_status_change_view(self):
        with self.admin_access.cnx() as cnx:
            u = self.create_user(cnx, "toto")
            iwfable = u.cw_adapt_to("IWorkflowable")
            iwfable.fire_transition("deactivate", comment="yeah")
            self.assertFalse(MAILBOX)
            cnx.commit()
            self.assertEqual(len(MAILBOX), 1)
            email = MAILBOX[0]
            self.assertEqual(
                email.content,
                f"""
admin changed status from <activated> to <deactivated> for entity
'toto'

yeah

url: {BASE_URL}cwuser/toto
""",
            )
            self.assertEqual(email.subject, f"status changed CWUser #{u.eid} (admin)")


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2020 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""Defines the central class for the CubicWeb RQL server: the repository.

The repository is an abstraction allowing execution of rql queries against
data sources. Most of the work is actually done in helper classes. The
repository mainly:

* brings these classes all together to provide a single access
  point to a cubicweb instance.
* handles session management
"""

import queue
import threading
import time
from contextlib import contextmanager
from itertools import chain
from logging import getLogger

from logilab.common.decorators import cached, clear_cache
from rql.utils import rqlvar_maker
from yams import BadSchemaDefinition

from cubicweb import (
    CW_MIGRATION_MAP,
    UnknownEid,
    AuthenticationError,
    ExecutionError,
    UniqueTogetherError,
    ViolatedConstraint,
)
from cubicweb import cwvreg, schema, server
from cubicweb import set_log_methods
from cubicweb.server import utils, hook, querier, sources
from cubicweb.server.session import InternalManager, Connection

NO_CACHE_RELATIONS = set(
    [
        ("owned_by", "object"),
        ("created_by", "object"),
        ("cw_source", "object"),
    ]
)


def prefill_entity_caches(entity):
    cnx = entity._cw
    # prefill entity relation caches
    for rschema in list(entity.e_schema.subject_relations.values()):
        rtype = str(rschema)
        if rtype in schema.VIRTUAL_RTYPES or (rtype, "subject") in NO_CACHE_RELATIONS:
            continue
        if rschema.final:
            entity.cw_attr_cache.setdefault(rtype, None)
        else:
            entity.cw_set_relation_cache(rtype, "subject", cnx.empty_rset())
    for rschema in list(entity.e_schema.object_relations.values()):
        rtype = str(rschema)
        if rtype in schema.VIRTUAL_RTYPES or (rtype, "object") in NO_CACHE_RELATIONS:
            continue
        entity.cw_set_relation_cache(rtype, "object", cnx.empty_rset())


def del_existing_rel_if_needed(cnx, eidfrom, rtype, eidto):
    """delete existing relation when adding a new one if card is 1 or ?

    have to be done once the new relation has been inserted to avoid having
    an entity without a relation for some time

    this kind of behaviour has to be done in the repository so we don't have
    hooks order hazardness
    """
    # skip that if integrity explicitly disabled
    if not cnx.is_hook_category_activated("activeintegrity"):
        return
    rdef = cnx.rtype_eids_rdef(rtype, eidfrom, eidto)
    card = rdef.cardinality
    # one may be tented to check for neweids but this may cause more than one
    # relation even with '1?'  cardinality if thoses relations are added in the
    # same transaction where the entity is being created. This never occurs from
    # the web interface but may occurs during test or dbapi connection (though
    # not expected for this).  So: don't do it, we pretend to ensure repository
    # consistency.
    #
    # notes:
    # * inlined relations will be implicitly deleted for the subject entity
    # * we don't want read permissions to be applied but we want delete
    #   permission to be checked
    if card[0] in "1?":
        with cnx.security_enabled(read=False):
            cnx.execute(
                "DELETE X %s Y WHERE X eid %%(x)s, " "NOT Y eid %%(y)s" % rtype,
                {"x": eidfrom, "y": eidto},
            )
    if card[1] in "1?":
        with cnx.security_enabled(read=False):
            cnx.execute(
                "DELETE X %s Y WHERE Y eid %%(y)s, " "NOT X eid %%(x)s" % rtype,
                {"x": eidfrom, "y": eidto},
            )


def preprocess_inlined_relations(cnx, entity):
    """when an entity is added, check if it has some inlined relation which
    requires to be extrated for proper call hooks
    """
    relations = []
    activeintegrity = cnx.is_hook_category_activated("activeintegrity")
    eschema = entity.e_schema
    for attr in entity.cw_edited:
        rschema = eschema.subject_relations[attr]
        if not rschema.final:  # inlined relation
            value = entity.cw_edited[attr]
            relations.append((attr, value))
            cnx.update_rel_cache_add(entity.eid, attr, value)
            rdef = cnx.rtype_eids_rdef(attr, entity.eid, value)
            if rdef.cardinality[1] in "1?" and activeintegrity:
                with cnx.security_enabled(read=False):
                    cnx.execute(
                        "DELETE X %s Y WHERE Y eid %%(y)s" % attr,
                        {"x": entity.eid, "y": value},
                    )
    return relations


class NullEventBus:
    def publish(self, msg):
        pass

    def add_subscription(self, topic, callback):
        pass

    def start(self):
        pass

    def stop(self):
        pass


# python STL doens't provide any concurrent deque implementation with blocking .get
class LifoDeque(queue.LifoQueue):
    def get_oldest(self, block=True, timeout=None):
        "Like queue.get except it gets the oldest one"
        with self.not_empty:
            if not block:
                if not self._qsize():
                    raise queue.Empty
            elif timeout is None:
                while not self._qsize():
                    self.not_empty.wait()
            elif timeout < 0:
                raise ValueError("'timeout' must be a non-negative number")
            else:
                endtime = time() + timeout
                while not self._qsize():
                    remaining = endtime - time()
                    if remaining <= 0.0:
                        raise queue.Empty
                    self.not_empty.wait(remaining)
            item = self.queue.pop(
                0
            )  # <- that's the only code difference from super().get
            self.not_full.notify()
        return item


class _BaseCnxSet:
    """
    Simple connection set without any pooling capability.

    You should only uses this if you are using an external pool like pgbouncer.
    """

    def __init__(self, source):
        self._source = source

    def qsize(self):
        return None

    def get(self):
        return self._new_cnxset()

    def _new_cnxset(self):
        return self._source.wrapped_connection()

    def release(self, cnxset):
        cnxset.close(True)

    def __iter__(self):
        return
        yield  # needed to be considered as a generator

    def close(self):
        pass


class _CnxSetPool(_BaseCnxSet):
    """
    Dynamic database connections pool.
    """

    def __init__(self, source, min_size=1, max_size=4, idle_timeout=300):
        """
        Connection pool for database connections. This pool is a dynamic pooler
        that opens new connections when the load is to high and closes connections
        on lower load after some idle timeout.

        Arguments::

        * source: database source
        * min_size: minimum number of simultanuous connections
        * max_size: maximum number of simultanuous connections, if set to 0
          there won't be any limit
        * idle_timeout: time before the pool starts closing connections
        """
        super().__init__(source)
        self._cnxsets = []
        self._queue = LifoDeque()
        self.lock = threading.Lock()
        self.min_size = min_size
        self.max_size = max_size
        self.idle = time.time()
        self.idle_timeout = idle_timeout

        for i in range(min_size):
            self._queue.put_nowait(self._new_cnxset())

    def _new_cnxset(self):
        """
        Create a new connection and returns it. This operation grabs the lock.
        """
        cnxset = super()._new_cnxset()
        with self.lock:
            self._cnxsets.append(cnxset)
        return cnxset

    def _close_idle_cnxset(self):
        """
        Close connections not being used since idle_timeout.
        """
        if (
            abs(time.time() - self.idle) > self.idle_timeout
            and self.size() > self.min_size
        ):
            try:
                cnxset = self._queue.get_oldest(block=False)
            except queue.Empty:
                # the queue has been used since we checked it size
                pass
            else:
                self.debug("[pool] load is low, close a connection")
                cnxset.close(True)
                with self.lock:
                    self._cnxsets.remove(cnxset)

    def size(self) -> int:
        """
        Return the total number of connections.
        """
        with self.lock:
            return len(self._cnxsets)

    def qsize(self) -> int:
        """
        Return the size of the queue.
        """
        return self._queue.qsize()

    def get(self):
        """
        Try to get a connection from the queue if available or open a new one
        if the maximum number isn't reached.

        If no connection is available and the maximum number of connections is
        reached, try to get a new one with a timeout. If the timeout is reached
        an exception is raised. The maximum number of connections can be
        modified to avoid that or the timeout can be increased.

        If max_size is set to 0 then there won't be any limit to open new connections.
        """
        self.debug(
            f"[pool] try to get a connection, free queue size is "
            f"{self.qsize()}/{self.size()} (max {self.max_size})"
        )
        try:
            cnxset = self._queue.get_nowait()
            self._close_idle_cnxset()
            self.debug("[pool] we got a directly available connection")
            return cnxset
        except queue.Empty:
            # reset idle time
            self.idle = time.time()
            if self.max_size and self.size() >= self.max_size:
                try:
                    connection = self._queue.get(True, timeout=5)
                    self.debug("[pool] we got a connection before minimum timeout")
                    return connection
                except queue.Empty:
                    raise Exception(
                        "no connections set available after 5 secs, probably either a "
                        "bug in code (too many uncommited/rolled back "
                        "connections) or too much load on the server (in "
                        "which case you can try to set a bigger connections pool size "
                        "by changing the connections-pool-max-size option in your "
                        "configuration file)"
                    )
            else:
                self.debug(
                    "[pool] we had to open a new connection, connection number: %d/%d"
                    % (self.size(), self.max_size)
                )
                return self._new_cnxset()

    def release(self, cnxset):
        """
        Release a connection by returning it into the queue.
        """
        self._queue.put_nowait(cnxset)
        self.debug(
            f"[pool] release {cnxset} [{id(cnxset)}], free queue size is "
            f"{self.qsize()}/{self.size()} (max {self.max_size})"
        )
        self._close_idle_cnxset()

    def __iter__(self):
        """
        Iter on all connections. This operation grabs the lock.
        """
        with self.lock:
            for cnxset in self._cnxsets:
                yield cnxset

    def close(self):
        """
        Close the poll by closing all the connections still in the queue.
        """
        while True:
            try:
                cnxset = self._queue.get_nowait()
            except queue.Empty:
                break
            try:
                cnxset.close(True)
            except Exception as e:
                self.exception(f"error while closing {cnxset}, error: {e}")

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


def get_cnxset(config, source, bootstrap=False):
    if not config["connections-pooler-enabled"]:
        return _BaseCnxSet(source)
    idle_timeout = config["connections-pool-idle-timeout"]
    if bootstrap or config.quick_start:
        min_size, max_size = 0, 1
    else:
        min_size = config["connections-pool-min-size"]
        max_size = config["connections-pool-max-size"]
    return _CnxSetPool(
        source, min_size=min_size, max_size=max_size, idle_timeout=idle_timeout
    )


class Repository:
    """a repository provides access to a set of persistent storages for
    entities and relations
    """

    def __init__(self, config, scheduler=None, vreg=None):
        self.config = config
        if vreg is None:
            vreg = cwvreg.CWRegistryStore(config)
        self.vreg = vreg
        self._scheduler = scheduler

        self.app_instances_bus = NullEventBus()

        # list of functions to be called at regular interval
        # list of running threads
        self._running_threads = []
        # initial schema, should be build or replaced latter
        self.schema = schema.CubicWebSchema(config.appid)
        self.vreg.schema = self.schema  # until actual schema is loaded...
        # shutdown flag
        self.shutting_down = None
        # sources (additional sources info in the system database)
        self.system_source = self.get_source(
            "native", "system", config.system_source_config.copy()
        )
        # querier helper, need to be created after sources initialization
        self.querier = querier.QuerierHelper(self, self.schema)
        # cache eid -> type
        self._type_cache = {}
        # the hooks manager
        self.hm = hook.HooksManager(self.vreg)

    def bootstrap(self):
        self.info("starting repository from %s", self.config.apphome)
        self.shutting_down = False
        config = self.config
        # 0. init a cnxset that will be used to fetch bootstrap information from
        #    the database
        self.cnxsets = get_cnxset(config, self.system_source, bootstrap=True)
        # 1. set used cubes
        if config.creating or not config.read_instance_schema:
            config.bootstrap_cubes()
        else:
            self.set_schema(self.config.load_bootstrap_schema(), resetvreg=False)
            config.init_cubes(self.get_cubes())
        # 2. load schema
        if config.quick_start:
            # quick start: only to get a minimal repository to get cubes
            # information (eg dump/restore/...)
            #
            # restrict appobject_path to only load hooks and entity classes in
            # the registry
            config.cube_appobject_path = set(("hooks", "entities"))
            config.cubicweb_appobject_path = set(("hooks", "entities"))
        if config.quick_start or config.creating or not config.read_instance_schema:
            # load schema from the file system
            if not config.creating:
                self.info("set fs instance's schema")
            self.set_schema(config.load_schema(expand_cubes=True))
            if not config.creating:
                # set eids on entities schema
                with self.internal_cnx() as cnx:
                    for etype, eid in cnx.execute(
                        "Any XN,X WHERE X is CWEType, X name XN"
                    ):
                        try:
                            self.schema.entity_schema_for(etype).eid = eid
                        except KeyError:
                            # etype in the database doesn't exist in the fs schema, this may occur
                            # during dev and we shouldn't crash
                            self.warning(
                                "No %s entity type in the file system schema", etype
                            )
        else:
            # normal start: load the instance schema from the database
            self.info("loading schema from the repository")
            self.set_schema(self.deserialize_schema())
        # 3. initialize data sources
        if config.creating:
            # call init_creating so that for instance native source can
            # configurate tsearch according to postgres version
            self.system_source.init_creating()
        else:
            self._init_system_source()
            if "CWProperty" in self.schema:
                self.vreg.init_properties(self.properties())
        # 4. close initialization connection set and reopen fresh ones for
        #    proper initialization
        self.cnxsets.close()
        self.cnxsets = get_cnxset(config, self.system_source)
        # 5. call instance level initialisation hooks
        self.hm.call_hooks("server_startup", repo=self)

    def source_by_uri(self, uri):
        with self.internal_cnx() as cnx:
            rset = cnx.find("CWSource", name=uri)
            if not rset:
                raise ValueError(f"no source with uri {uri} found")
            return self._source_from_cwsource(rset.one())

    def source_by_eid(self, eid):
        with self.internal_cnx() as cnx:
            rset = cnx.find("CWSource", eid=eid)
            if not rset:
                raise ValueError("no source with eid %d found" % eid)
            return self._source_from_cwsource(rset.one())

    @property
    def sources_by_uri(self):
        mapping = {"system": self.system_source}
        mapping.update(
            (sourceent.name, source) for sourceent, source in self._sources()
        )
        return mapping

    def _sources(self):
        if self.config.quick_start:
            return
        with self.internal_cnx() as cnx:
            for sourceent in cnx.execute(
                "Any S, SN, SA, SC WHERE S is_instance_of CWSource, "
                'S name SN, S type SA, S config SC, S name != "system"'
            ).entities():
                source = self._source_from_cwsource(sourceent)
                yield sourceent, source
        self._clear_source_defs_caches()

    def _source_from_cwsource(self, sourceent):
        source = self.get_source(
            sourceent.type, sourceent.name, sourceent.host_config, sourceent.eid
        )
        if self.config.source_enabled(source):
            # call source's init method to complete their initialisation if
            # needed (for instance looking for persistent configuration using an
            # internal session, which is not possible until connections sets have been
            # initialized)
            source.init(sourceent)
        return source

    # internals ###############################################################

    def _init_system_source(self):
        if self.config.quick_start:
            self.system_source.init_creating()
            return
        with self.internal_cnx() as cnx:
            sourceent = cnx.execute(
                "Any S, SA, SC WHERE S is_instance_of CWSource,"
                ' S name "system", S type SA, S config SC'
            ).one()
            self.system_source.eid = sourceent.eid
            self.system_source.init(sourceent)

    def get_source(self, type, uri, source_config, eid=None):
        # set uri and type in source config so it's available through
        # source_defs()
        source_config["uri"] = uri
        source_config["type"] = type
        return sources.get_source(type, source_config, self, eid)

    def set_schema(self, schema, resetvreg=True):
        self.info("set schema %s %#x", schema.name, id(schema))
        if resetvreg:
            # trigger full reload of all appobjects
            self.vreg.set_schema(schema)
        else:
            self.vreg._set_schema(schema)
        self.querier.set_schema(schema)
        self.system_source.set_schema(schema)
        self.schema = schema

    def deserialize_schema(self):
        """load schema from the database"""
        from cubicweb.server.schemaserial import deserialize_schema

        appschema = schema.CubicWebSchema(self.config.appid)
        self.debug("deserializing db schema into %s %#x", appschema.name, id(appschema))
        with self.internal_cnx() as cnx:
            try:
                deserialize_schema(appschema, cnx)
            except BadSchemaDefinition:
                raise
            except Exception as ex:
                import traceback

                traceback.print_exc()
                raise Exception(f"Is the database initialised ? (cause: {ex})")
        return appschema

    def has_scheduler(self) -> bool:
        """Return True if the repository has a scheduler attached and is able
        to register looping tasks.
        """
        return self._scheduler is not None

    def run_scheduler(self):
        """Start repository scheduler after preparing the repository for that.

        * trigger server startup hook,
        * start the scheduler *and block*.

        XXX Other startup related stuffs are done elsewhere. In Repository
        XXX __init__ or in external codes (various server managers).
        """
        assert (
            self.has_scheduler()
        ), "This Repository is not intended to be used as a server"
        self.info(
            "starting repository scheduler with tasks: %s",
            ", ".join(e.action.__name__ for e in self._scheduler.queue),
        )
        self._scheduler.run()

    def looping_task(self, interval, func, *args):
        """register a function to be called every `interval` seconds.

        looping tasks can only be registered during repository initialization,
        once done this method will fail.
        """
        if self.config.repairing:
            return
        if not self.has_scheduler():
            self.warning(
                "looping task %s will not run in this process where repository "
                'has no scheduler; use "cubicweb-ctl scheduler <appid>" to '
                "have it running",
                func,
            )
            return
        event = utils.schedule_periodic_task(self._scheduler, interval, func, *args)
        self.info(
            "scheduled periodic task %s (interval: %.2fs)",
            event.action.__name__,
            interval,
        )

    def threaded_task(self, func):
        """start function in a separated thread"""
        utils.RepoThread(func, self._running_threads).start()

    def shutdown(self):
        """called on server stop event to properly close opened sessions and
        connections
        """
        assert not self.shutting_down, "already shutting down"
        if not (
            self.config.creating or self.config.repairing or self.config.quick_start
        ):
            # then, the system source is still available
            self.hm.call_hooks("before_server_shutdown", repo=self)
        self.shutting_down = True
        self.info("shutting down repository")
        self.system_source.shutdown()
        if not (
            self.config.creating or self.config.repairing or self.config.quick_start
        ):
            self.hm.call_hooks("server_shutdown", repo=self)
        for thread in self._running_threads:
            self.info("waiting thread %s...", thread.getName())
            thread.join()
            self.info("thread %s finished", thread.getName())
        self.cnxsets.close()
        hits, misses = (
            self.querier.rql_cache.cache_hit,
            self.querier.rql_cache.cache_miss,
        )
        try:
            self.info(
                "rql st cache hit/miss: %s/%s (%s%% hits)",
                hits,
                misses,
                (hits * 100) / (hits + misses),
            )
            hits, misses = self.system_source.cache_hit, self.system_source.cache_miss
            self.info(
                "sql cache hit/miss: %s/%s (%s%% hits)",
                hits,
                misses,
                (hits * 100) / (hits + misses),
            )
            nocache = self.system_source.no_cache
            self.info(
                "sql cache usage: %s/%s (%s%%)",
                hits + misses,
                nocache,
                ((hits + misses) * 100) / (hits + misses + nocache),
            )
        except ZeroDivisionError:
            pass

    def check_auth_info(self, cnx, login, authinfo):
        """validate authentication, raise AuthenticationError on failure, return
        associated CWUser's eid on success.
        """
        # iter on sources_by_uri then check enabled source since sources doesn't
        # contain copy based sources
        for source in self.sources_by_uri.values():
            if self.config.source_enabled(source):
                try:
                    auth = source.authenticate(cnx, login, **authinfo)
                    return auth
                except (NotImplementedError, AuthenticationError):
                    continue
        else:
            raise AuthenticationError(
                f"authentication failed with all sources "
                f"({self.sources_by_uri.values()})"
            )

    def authenticate_user(self, cnx, login, **authinfo):
        """validate login / password, raise AuthenticationError on failure
        return associated CWUser instance on success
        """
        eid = self.check_auth_info(cnx, login, authinfo)
        cwuser = self._build_user(cnx, eid)
        if (
            self.config.consider_user_state
            and cwuser.cw_adapt_to("IWorkflowable").state
            not in cwuser.AUTHENTICABLE_STATES
        ):
            # TODO use f'{a=}' once we get to python 3.8
            raise AuthenticationError(
                f"user is not in authenticable state (self.config.consider_user_state="
                f"{self.config.consider_user_state} and cwuser.cw_adapt_to('IWorkflowable').state="
                f"{cwuser.cw_adapt_to('IWorkflowable').state} not in cwuser.AUTHENTICABLE_STATES="
                f"{cwuser.AUTHENTICABLE_STATES}"
            )
        return cwuser

    def _build_user(self, cnx, eid):
        """return a CWUser entity for user with the given eid"""
        cls = self.vreg["etypes"].etype_class("CWUser")
        st = cls.fetch_rqlst(cnx.user, ordermethod=None)
        st.add_eid_restriction(st.get_variable("X"), "x", "Substitute")
        rset = cnx.execute(st.as_string(), {"x": eid})
        assert len(rset) == 1, rset
        return rset.get_entity(0, 0)

    # public (dbapi) interface ################################################

    def get_schema(self):
        """Return the instance schema.

        This is a public method, not requiring a session id.
        """
        return self.schema

    def get_cubes(self):
        """Return the list of cubes used by this instance.

        This is a public method, not requiring a session id.
        """
        versions = self.get_versions(
            not (
                self.config.creating
                or self.config.repairing
                or self.config.quick_start
                or self.config.mode == "test"
            )
        )
        cubes = list(versions)
        cubes.remove("cubicweb")
        return cubes

    def get_option_value(self, option):
        """Return the value for `option` in the configuration.

        This is a public method, not requiring a session id.
        """
        # XXX we may want to check we don't give sensible information
        return self.config[option]

    @cached
    def get_versions(self, checkversions=False):
        """Return the a dictionary containing cubes used by this instance
        as key with their version as value, including cubicweb version.

        This is a public method, not requiring a session id.
        """
        from logilab.common.changelog import Version

        vcconf = {}
        with self.internal_cnx() as cnx:
            for pk, version in cnx.execute(
                "Any K,V WHERE P is CWProperty, P value V, P pkey K, "
                'P pkey ~="system.version.%"',
                build_descr=False,
            ):
                cube = pk.split(".")[-1]
                # XXX cubicweb migration
                if cube in CW_MIGRATION_MAP:
                    cube = CW_MIGRATION_MAP[cube]
                version = Version(version)
                vcconf[cube] = version
                if checkversions:
                    if cube != "cubicweb":
                        fsversion = self.config.cube_version(cube)
                    else:
                        fsversion = self.config.cubicweb_version()
                    if version < fsversion:
                        msg = (
                            "instance has %s version %s but %s "
                            'is installed. Run "cubicweb-ctl upgrade %s".'
                        )
                        raise ExecutionError(
                            msg % (cube, version, fsversion, self.config.appid)
                        )
        return vcconf

    @cached
    def source_defs(self):
        """Return the a dictionary containing source uris as value and a
        dictionary describing each source as value.

        This is a public method, not requiring a session id.
        """
        sources = {}
        # remove sensitive information
        for uri, source in self.sources_by_uri.items():
            sources[uri] = source.public_config
        return sources

    def _clear_source_defs_caches(self):
        clear_cache(self, "source_defs")

    def properties(self):
        """Return a result set containing system wide properties.

        This is a public method, not requiring a session id.
        """
        with self.internal_cnx() as cnx:
            # don't use cnx.execute, we don't want rset.req set
            return self.querier.execute(
                cnx,
                "Any K,V WHERE P is CWProperty,"
                "P pkey K, P value V, NOT P for_user U",
                build_descr=False,
            )

    def find_users(self, fetch_attrs, **query_attrs):
        """yield user attributes for cwusers matching the given query_attrs
        (the result set cannot survive this method call)

        This can be used by low-privileges account (anonymous comes to
        mind).

        `fetch_attrs`: tuple of attributes to be fetched
        `query_attrs`: dict of attr/values to restrict the query
        """
        assert query_attrs
        if not hasattr(self, "_cwuser_attrs"):
            cwuser = self.schema["CWUser"]
            self._cwuser_attrs = set(
                str(rschema)
                for rschema, _eschema in cwuser.attribute_definitions()
                if not rschema.meta
            )
        cwuserattrs = self._cwuser_attrs
        for k in chain(fetch_attrs, query_attrs):
            if k not in cwuserattrs:
                raise Exception("bad input for find_user")
        with self.internal_cnx() as cnx:
            varmaker = rqlvar_maker()
            vars = [(attr, next(varmaker)) for attr in fetch_attrs]
            rql = f"Any {','.join(var[1] for var in vars)} WHERE X is CWUser, "
            rql += ",".join(f"X {var[0]} {var[1]}" for var in vars) + ","
            rset = cnx.execute(
                rql + ",".join(f"X {attr} %({attr})s" for attr in query_attrs),
                query_attrs,
            )
            return rset.rows

    # session handling ########################################################

    @contextmanager
    def internal_cnx(self):
        """Context manager returning a Connection using internal user which have
        every access rights on the repository.

        Internal connections have all hooks beside security enabled.
        """
        with Connection(self, InternalManager()) as cnx:
            cnx.user._cw = (
                cnx  # XXX remove when "vreg = user._cw.vreg" hack in entity.py is gone
            )
            with cnx.security_enabled(read=False, write=False):
                yield cnx

    # data sources handling ###################################################
    # * correspondance between eid and type
    # * correspondance between eid and local id (i.e. specific to a given source)

    def clear_caches(self, eids=None):
        if eids is None:
            self._type_cache = {}
            etypes = None
        else:
            etypes = []
            etcache = self._type_cache
            for eid in eids:
                try:
                    etype = etcache.pop(int(eid))  # may be a string in some cases
                except KeyError:
                    etype = None
                etypes.append(etype)
        self.querier.clear_caches(eids, etypes)
        self.system_source.clear_caches(eids, etypes)

    def type_from_eid(self, eid, cnx):
        """Return the type of the entity with id `eid`"""
        try:
            eid = int(eid)
        except ValueError:
            raise UnknownEid(eid)
        try:
            return self._type_cache[eid]
        except KeyError:
            etype = self.system_source.eid_type(cnx, eid)
            self._type_cache[eid] = etype
            return etype

    def add_info(self, cnx, entity, source):
        """add type and source info for an eid into the system table,
        and index the entity with the full text index
        """
        # begin by inserting eid/type/source into the entities table
        hook.CleanupNewEidsCacheOp.get_instance(cnx).add_data(entity.eid)
        self.system_source.add_info(cnx, entity, source)

    def _delete_cascade_multi(self, cnx, entities):
        """same as _delete_cascade but accepts a list of entities with
        the same etype and belonging to the same source.
        """
        pendingrtypes = cnx.transaction_data.get("pendingrtypes", ())
        # delete remaining relations: if user can delete the entity, he can
        # delete all its relations without security checking
        with cnx.security_enabled(read=False, write=False):
            in_eids = ",".join([str(_e.eid) for _e in entities])
            with cnx.running_hooks_ops():
                for rschema, _, role in entities[0].e_schema.relation_definitions():
                    if rschema.rule:
                        continue  # computed relation
                    rtype = rschema.type
                    if rtype in schema.VIRTUAL_RTYPES or rtype in pendingrtypes:
                        continue
                    if role == "subject":
                        # don't skip inlined relation so they are regularly
                        # deleted and so hooks are correctly called
                        rql = f"DELETE X {rtype} Y WHERE X eid IN ({in_eids})"
                    else:
                        rql = f"DELETE Y {rtype} X WHERE X eid IN ({in_eids})"
                    cnx.execute(rql, build_descr=False)

    def init_entity_caches(self, cnx, entity, source):
        """Add entity to connection entities cache and repo's cache."""
        cnx.set_entity_cache(entity)
        self._type_cache[entity.eid] = entity.cw_etype

    def glob_add_entity(self, cnx, edited):
        """add an entity to the repository

        the entity eid should originally be None and a unique eid is assigned to
        the entity instance
        """
        entity = edited.entity
        entity._cw_is_saved = False  # entity has an eid but is not yet saved
        # init edited_attributes before calling before_add_entity hooks
        entity.cw_edited = edited
        source = self.system_source
        # allocate an eid to the entity before calling hooks
        entity.eid = self.system_source.create_eid(cnx)
        # set caches asap
        self.init_entity_caches(cnx, entity, source)
        if server.DEBUG & server.DBG_REPO:
            print("ADD entity", self, entity.cw_etype, entity.eid, edited)
        prefill_entity_caches(entity)
        self.hm.call_hooks("before_add_entity", cnx, entity=entity)
        relations = preprocess_inlined_relations(cnx, entity)
        edited.set_defaults()
        if cnx.is_hook_category_activated("integrity"):
            edited.check(creation=True)
        self.add_info(cnx, entity, source)
        try:
            source.add_entity(cnx, entity)
        except (UniqueTogetherError, ViolatedConstraint) as exc:
            userhdlr = cnx.vreg["adapters"].select(
                "IUserFriendlyError", cnx, entity=entity, exc=exc
            )
            userhdlr.raise_user_exception()
        edited.saved = entity._cw_is_saved = True
        # trigger after_add_entity after after_add_relation
        self.hm.call_hooks("after_add_entity", cnx, entity=entity)
        # call hooks for inlined relations
        for attr, value in relations:
            self.hm.call_hooks(
                "before_add_relation", cnx, eidfrom=entity.eid, rtype=attr, eidto=value
            )
            self.hm.call_hooks(
                "after_add_relation", cnx, eidfrom=entity.eid, rtype=attr, eidto=value
            )
        return entity.eid

    def glob_update_entity(self, cnx, edited):
        """replace an entity in the repository
        the type and the eid of an entity must not be changed
        """
        entity = edited.entity
        if server.DEBUG & server.DBG_REPO:
            print(
                "UPDATE entity",
                entity.cw_etype,
                entity.eid,
                entity.cw_attr_cache,
                edited,
            )
        hm = self.hm
        eschema = entity.e_schema
        cnx.set_entity_cache(entity)
        orig_edited = getattr(entity, "cw_edited", None)
        entity.cw_edited = edited
        source = self.system_source
        try:
            only_inline_rels, need_fti_update = True, False
            relations = []
            for attr in list(edited):
                if attr == "eid":
                    continue
                rschema = eschema.subject_relations[attr]
                if rschema.final:
                    if getattr(
                        eschema.relation_definition(attr), "fulltextindexed", False
                    ):
                        need_fti_update = True
                    only_inline_rels = False
                else:
                    # inlined relation
                    previous_value = entity.related(attr) or None
                    if previous_value is not None:
                        previous_value = previous_value[0][0]  # got a result set
                        if previous_value == entity.cw_attr_cache[attr]:
                            previous_value = None
                        else:
                            hm.call_hooks(
                                "before_delete_relation",
                                cnx,
                                eidfrom=entity.eid,
                                rtype=attr,
                                eidto=previous_value,
                            )
                    relations.append((attr, edited[attr], previous_value))
            # call hooks for inlined relations
            for attr, value, _t in relations:
                hm.call_hooks(
                    "before_add_relation",
                    cnx,
                    eidfrom=entity.eid,
                    rtype=attr,
                    eidto=value,
                )
            if not only_inline_rels:
                hm.call_hooks("before_update_entity", cnx, entity=entity)
            if cnx.is_hook_category_activated("integrity"):
                edited.check()
            try:
                source.update_entity(cnx, entity)
                edited.saved = True
            except (UniqueTogetherError, ViolatedConstraint) as exc:
                userhdlr = cnx.vreg["adapters"].select(
                    "IUserFriendlyError", cnx, entity=entity, exc=exc
                )
                userhdlr.raise_user_exception()
            self.system_source.update_info(cnx, entity, need_fti_update)
            if not only_inline_rels:
                hm.call_hooks("after_update_entity", cnx, entity=entity)
            for attr, value, prevvalue in relations:
                # if the relation is already cached, update existant cache
                relcache = entity.cw_relation_cached(attr, "subject")
                if prevvalue is not None:
                    hm.call_hooks(
                        "after_delete_relation",
                        cnx,
                        eidfrom=entity.eid,
                        rtype=attr,
                        eidto=prevvalue,
                    )
                    if relcache is not None:
                        cnx.update_rel_cache_del(entity.eid, attr, prevvalue)
                del_existing_rel_if_needed(cnx, entity.eid, attr, value)
                cnx.update_rel_cache_add(entity.eid, attr, value)
                hm.call_hooks(
                    "after_add_relation",
                    cnx,
                    eidfrom=entity.eid,
                    rtype=attr,
                    eidto=value,
                )
        finally:
            if orig_edited is not None:
                entity.cw_edited = orig_edited

    def glob_delete_entities(self, cnx, eids):
        """delete a list of  entities and all related entities from the repository"""
        # mark eids as being deleted in cnx info and setup cache update
        # operation (register pending eids before actual deletion to avoid
        # multiple call to glob_delete_entities)
        op = hook.CleanupDeletedEidsCacheOp.get_instance(cnx)
        eids = eids - op._container
        op._container |= eids
        data_by_etype = {}  # values are [list of entities]
        #
        # WARNING: the way this dictionary is populated is heavily optimized
        # and does not use setdefault on purpose. Unless a new release
        # of the Python interpreter advertises large perf improvements
        # in setdefault, this should not be changed without profiling.
        for eid in eids:
            etype = self.type_from_eid(eid, cnx)
            entity = cnx.entity_from_eid(eid, etype)
            try:
                data_by_etype[etype].append(entity)
            except KeyError:
                data_by_etype[etype] = [entity]
        source = self.system_source
        for etype, entities in data_by_etype.items():
            if server.DEBUG & server.DBG_REPO:
                print("DELETE entities", etype, [e.eid for e in entities])
            self.hm.call_hooks("before_delete_entity", cnx, entities=entities)
            self._delete_cascade_multi(cnx, entities)
            source.delete_entities(cnx, entities)
            source.delete_info_multi(cnx, entities)
            self.hm.call_hooks("after_delete_entity", cnx, entities=entities)
        # don't clear cache here, it is done in a hook on commit

    def glob_add_relation(self, cnx, subject, rtype, object):
        """add a relation to the repository"""
        self.glob_add_relations(cnx, {rtype: [(subject, object)]})

    def glob_add_relations(self, cnx, relations):
        """add several relations to the repository

        relations is a dictionary rtype: [(subj_eid, obj_eid), ...]
        """
        source = self.system_source
        relations_by_rtype = {}
        subjects_by_types = {}
        objects_by_types = {}
        activintegrity = cnx.is_hook_category_activated("activeintegrity")
        for rtype, eids_subj_obj in relations.items():
            if server.DEBUG & server.DBG_REPO:
                for subjeid, objeid in eids_subj_obj:
                    print("ADD relation", subjeid, rtype, objeid)
            for subjeid, objeid in eids_subj_obj:
                relations_by_rtype.setdefault(rtype, []).append((subjeid, objeid))
                if not activintegrity:
                    continue
                # take care to relation of cardinality '?1', as all eids will
                # be inserted later, we've remove duplicated eids since they
                # won't be caught by `del_existing_rel_if_needed`
                rdef = cnx.rtype_eids_rdef(rtype, subjeid, objeid)
                card = rdef.cardinality
                if card[0] in "?1":
                    with cnx.security_enabled(read=False):
                        cnx.execute(
                            "DELETE X %s Y WHERE X eid %%(x)s, "
                            "NOT Y eid %%(y)s" % rtype,
                            {"x": subjeid, "y": objeid},
                        )
                    subjects = subjects_by_types.setdefault(rdef, {})
                    if subjeid in subjects:
                        del relations_by_rtype[rtype][subjects[subjeid]]
                        subjects[subjeid] = len(relations_by_rtype[rtype]) - 1
                        continue
                    subjects[subjeid] = len(relations_by_rtype[rtype]) - 1
                if card[1] in "?1":
                    with cnx.security_enabled(read=False):
                        cnx.execute(
                            "DELETE X %s Y WHERE Y eid %%(y)s, "
                            "NOT X eid %%(x)s" % rtype,
                            {"x": subjeid, "y": objeid},
                        )
                    objects = objects_by_types.setdefault(rdef, {})
                    if objeid in objects:
                        del relations_by_rtype[rtype][objects[objeid]]
                        objects[objeid] = len(relations_by_rtype[rtype])
                        continue
                    objects[objeid] = len(relations_by_rtype[rtype])
        for rtype, source_relations in relations_by_rtype.items():
            self.hm.call_hooks(
                "before_add_relation", cnx, rtype=rtype, eids_from_to=source_relations
            )
        for rtype, source_relations in relations_by_rtype.items():
            source.add_relations(cnx, rtype, source_relations)
            rschema = self.schema.relation_schema_for(rtype)
            for subjeid, objeid in source_relations:
                cnx.update_rel_cache_add(subjeid, rtype, objeid, rschema.symmetric)
        for rtype, source_relations in relations_by_rtype.items():
            self.hm.call_hooks(
                "after_add_relation", cnx, rtype=rtype, eids_from_to=source_relations
            )

    def glob_delete_relation(self, cnx, subject, rtype, object):
        """delete a relation from the repository"""
        if server.DEBUG & server.DBG_REPO:
            print("DELETE relation", subject, rtype, object)
        source = self.system_source
        self.hm.call_hooks(
            "before_delete_relation", cnx, eidfrom=subject, rtype=rtype, eidto=object
        )
        source.delete_relation(cnx, subject, rtype, object)
        rschema = self.schema.relation_schema_for(rtype)
        cnx.update_rel_cache_del(subject, rtype, object, rschema.symmetric)
        self.hm.call_hooks(
            "after_delete_relation", cnx, eidfrom=subject, rtype=rtype, eidto=object
        )

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


set_log_methods(Repository, getLogger("cubicweb.repository"))
set_log_methods(_CnxSetPool, getLogger("cubicweb.repository.pool"))



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

import gettext
import locale

from logilab.common.deprecation import argument_removed


class cwGNUTranslations(gettext.GNUTranslations):
    # The encoding of a msgctxt and a msgid in a .mo file is
    # msgctxt + "\x04" + msgid (gettext version >= 0.15)
    CONTEXT_ENCODING = "%s\x04%s"

    def pgettext(self, context, message):
        ctxt_msg_id = self.CONTEXT_ENCODING % (context, message)
        missing = object()
        tmsg = self._catalog.get(ctxt_msg_id, missing)
        if tmsg is missing:
            if self._fallback:
                return self._fallback.pgettext(context, message)
            return message
        # Encode the Unicode tmsg back to an 8-bit string, if possible
        if self._output_charset:
            return tmsg.encode(self._output_charset)
        elif self._charset:
            return tmsg.encode(self._charset)
        return tmsg

    def lpgettext(self, context, message):
        ctxt_msg_id = self.CONTEXT_ENCODING % (context, message)
        missing = object()
        tmsg = self._catalog.get(ctxt_msg_id, missing)
        if tmsg is missing:
            if self._fallback:
                return self._fallback.lpgettext(context, message)
            return message
        if self._output_charset:
            return tmsg.encode(self._output_charset)
        return tmsg.encode(locale.getpreferredencoding())

    def npgettext(self, context, msgid1, msgid2, n):
        ctxt_msg_id = self.CONTEXT_ENCODING % (context, msgid1)
        try:
            tmsg = self._catalog[(ctxt_msg_id, self.plural(n))]
            if self._output_charset:
                return tmsg.encode(self._output_charset)
            elif self._charset:
                return tmsg.encode(self._charset)
            return tmsg
        except KeyError:
            if self._fallback:
                return self._fallback.npgettext(context, msgid1, msgid2, n)
            if n == 1:
                return msgid1
            else:
                return msgid2

    def lnpgettext(self, context, msgid1, msgid2, n):
        ctxt_msg_id = self.CONTEXT_ENCODING % (context, msgid1)
        try:
            tmsg = self._catalog[(ctxt_msg_id, self.plural(n))]
            if self._output_charset:
                return tmsg.encode(self._output_charset)
            return tmsg.encode(locale.getpreferredencoding())
        except KeyError:
            if self._fallback:
                return self._fallback.lnpgettext(context, msgid1, msgid2, n)
            if n == 1:
                return msgid1
            else:
                return msgid2

    ugettext = gettext.GNUTranslations.gettext

    def upgettext(self, context, message):
        ctxt_message_id = self.CONTEXT_ENCODING % (context, message)
        missing = object()
        tmsg = self._catalog.get(ctxt_message_id, missing)
        if tmsg is missing:
            # XXX logilab patch for compat w/ catalog generated by cw < 3.5
            return self.ugettext(message)
            if self._fallback:
                return self._fallback.upgettext(context, message)
            return str(message)
        return tmsg

    def unpgettext(self, context, msgid1, msgid2, n):
        ctxt_message_id = self.CONTEXT_ENCODING % (context, msgid1)
        try:
            tmsg = self._catalog[(ctxt_message_id, self.plural(n))]
        except KeyError:
            if self._fallback:
                return self._fallback.unpgettext(context, msgid1, msgid2, n)
            if n == 1:
                tmsg = str(msgid1)
            else:
                tmsg = str(msgid2)
        return tmsg


# https://docs.python.org/3/library/gettext.html#gettext.translation
# deprecated in python 3.8, removed in 3.10
@argument_removed("codeset", version="3.32")
def translation(
    domain, localedir=None, languages=None, class_=None, fallback=False, codeset=None
):
    if class_ is None:
        class_ = cwGNUTranslations
    return gettext.translation(
        domain,
        localedir=localedir,
        languages=languages,
        class_=class_,
        fallback=fallback,
    )



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.server.sqlutils
"""

from logilab.common.testlib import TestCase, unittest_main

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.sqlutils import SQLAdapterMixIn

BASE_CONFIG = {
    "db-driver": "Postgres",
    "db-host": "crater",
    "db-name": "cubicweb2_test",
    "db-user": "toto",
    "db-upassword": "toto",
}


class SQLAdapterMixInTC(TestCase):
    def test_init(self):
        o = SQLAdapterMixIn(BASE_CONFIG)
        self.assertEqual(o.dbhelper.dbencoding, "UTF-8")

    def test_init_encoding(self):
        config = BASE_CONFIG.copy()
        config["db-encoding"] = "ISO-8859-1"
        o = SQLAdapterMixIn(config)
        self.assertEqual(o.dbhelper.dbencoding, "ISO-8859-1")


class SQLUtilsTC(CubicWebTC):
    def test_group_concat(self):
        with self.admin_access.repo_cnx() as cnx:
            g = cnx.create_entity("CWGroup", name="héhé")
            cnx.create_entity("CWUser", login="toto", upassword="", in_group=g.eid)
            rset = cnx.execute(
                "Any L,GROUP_CONCAT(G) GROUPBY L WHERE X login L,"
                'X in_group G, G name GN, NOT G name IN ("users", "héhé")'
            )
            self.assertEqual([["admin", "3"], ["anon", "2"]], rset.rows)
            rset = cnx.execute(
                "Any L,GROUP_CONCAT(GN) GROUPBY L WHERE X login L,"
                'X in_group G, G name GN, NOT G name "users"'
            )
            self.assertEqual(
                [["admin", "managers"], ["anon", "guests"], ["toto", "héhé"]], rset.rows
            )


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Helper classes to execute RQL queries on a set of sources, performing
security checking and data aggregation.
"""
import time
import traceback
import uuid
from itertools import repeat
from logging import getLogger
from threading import Lock
from copy import deepcopy

from cachetools import LFUCache
from rql import RQLSyntaxError, CoercionError, BadRQLQuery
from rql.nodes import (
    ETYPE_PYOBJ_MAP,
    etype_from_pyobj,
    Relation,
    Exists,
    Not,
    VariableRef,
    VariableRefMethodCall,
    VariableRefAttributeAccess,
    Constant,
)
from rql.stmts import Union
from yams import BASE_TYPES

from cubicweb import Binary, server
from yams import ValidationError
from cubicweb import Unauthorized, UnknownEid, QueryError
from cubicweb import set_log_methods
from cubicweb.debug import debug_channel_has_subscribers, emit_to_debug_channel
from cubicweb.misc.source_highlight import highlight_terminal
from cubicweb.rqlrewrite import RQLRelationRewriter
from cubicweb.rset import ResultSet
from cubicweb.server.edition import EditedEntity
from cubicweb.server.rqlannotation import RQLAnnotator, set_qdata
from cubicweb.server.ssplanner import (
    READ_ONLY_RTYPES,
    add_types_restriction,
    prepare_plan,
)
from cubicweb.utils import RepeatList

ETYPE_PYOBJ_MAP[Binary] = "Bytes"


def empty_rset(rql, args):
    """build an empty result set object"""
    return ResultSet([], rql, args)


# permission utilities ########################################################


def check_no_password_selected(rqlst):
    """check that Password entities are not selected"""
    for solution in rqlst.solutions:
        for var, etype in solution.items():
            if etype == "Password":
                raise Unauthorized(f"Password selection is not allowed ({var})")


def term_etype(cnx, term, solution, args):
    """return the entity type for the given term (a VariableRef or a Constant
    node)
    """
    try:
        return solution[term.name]
    except AttributeError:
        return cnx.entity_type(term.eval(args))


def check_relations_read_access(cnx, select, args):
    """Raise :exc:`Unauthorized` if the given user doesn't have credentials to
    read relations used in the given syntax tree
    """
    # use `term_etype` since we've to deal with rewritten constants here,
    # when used as an external source by another repository.
    # XXX what about local read security w/ those rewritten constants...
    # XXX constants can also happen in some queries generated by req.find()
    DBG = (server.DEBUG & server.DBG_SEC) and "read" in server._SECURITY_CAPS
    schema = cnx.repo.schema
    user = cnx.user
    if select.where is not None:
        for rel in select.where.iget_nodes(Relation):
            for solution in select.solutions:
                # XXX has_text may have specific perm ?
                if rel.r_type in READ_ONLY_RTYPES:
                    continue
                rschema = schema.relation_schema_for(rel.r_type)
                if rschema.final:
                    eschema = schema.entity_schema_for(
                        term_etype(cnx, rel.children[0], solution, args)
                    )
                    rdef = eschema.relation_definition(rschema)
                else:
                    rdef = rschema.relation_definition(
                        term_etype(cnx, rel.children[0], solution, args),
                        term_etype(cnx, rel.children[1].children[0], solution, args),
                    )
                if not user.matching_groups(rdef.get_groups("read")):
                    if DBG:
                        print(
                            "check_read_access: %s %s does not match %s"
                            % (rdef, user.groups, rdef.get_groups("read"))
                        )
                    # XXX rqlexpr not allowed
                    raise Unauthorized("read", rel.r_type)
                if DBG:
                    print(
                        "check_read_access: %s %s matches %s"
                        % (rdef, user.groups, rdef.get_groups("read"))
                    )


def get_local_checks(cnx, rqlst, solution):
    """Check that the given user has credentials to access data read by the
    query and return a dict defining necessary "local checks" (i.e. rql
    expression in read permission defined in the schema) where no group grants
    him the permission.

    Returned dictionary's keys are variable names and values the rql expressions
    for this variable (with the given solution).

    Raise :exc:`Unauthorized` if access is known to be defined, i.e. if there is
    no matching group and no local permissions.
    """
    DBG = (server.DEBUG & server.DBG_SEC) and "read" in server._SECURITY_CAPS
    schema = cnx.repo.schema
    user = cnx.user
    localchecks = {}
    # iterate on defined_vars and not on solutions to ignore column aliases
    for varname in rqlst.defined_vars:
        eschema = schema.entity_schema_for(solution[varname])
        if eschema.final:
            continue
        if not user.matching_groups(eschema.get_groups("read")):
            erqlexprs = eschema.get_rqlexprs("read")
            if not erqlexprs:
                ex = Unauthorized("read", solution[varname])
                ex.var = varname
                if DBG:
                    print(
                        "check_read_access: %s %s %s %s"
                        % (varname, eschema, user.groups, eschema.get_groups("read"))
                    )
                raise ex
            # don't insert security on variable only referenced by 'NOT X relation Y' or
            # 'NOT EXISTS(X relation Y)'
            varinfo = rqlst.defined_vars[varname].stinfo
            if varinfo["selected"] or (
                len(
                    [
                        r
                        for r in varinfo["relations"]
                        if (
                            not schema.relation_schema_for(r.r_type).final
                            and (
                                (
                                    isinstance(r.parent, Exists)
                                    and r.parent.neged(strict=True)
                                )
                                or isinstance(r.parent, Not)
                            )
                        )
                    ]
                )
                != len(varinfo["relations"])
            ):
                localchecks[varname] = erqlexprs
    return localchecks


# Plans #######################################################################


class ExecutionPlan:
    """the execution model of a rql query, composed of querier steps"""

    def __init__(self, schema, rqlst, args, cnx):
        self.schema = schema
        # original rql syntax tree
        self.rqlst = rqlst
        self.args = args or {}
        # cnx executing the query
        self.cnx = cnx
        # execution steps
        self.steps = []
        # tracing token for debugging
        self.rql_query_tracing_token = None

    def add_step(self, step):
        """add a step to the plan"""
        self.steps.append(step)

    def sqlexec(self, sql, args=None):
        return self.cnx.repo.system_source.sqlexec(self.cnx, sql, args)

    def execute(self):
        """execute a plan and return resulting rows"""
        for step in self.steps:
            step.rql_query_tracing_token = self.rql_query_tracing_token
            result = step.execute()
        # the latest executed step contains the full query result
        return result

    def preprocess(self, union, security=True):
        """insert security when necessary then annotate rql syntax tree
        to prepare sql generation
        """
        cached = None
        if security and self.cnx.read_security:
            # ensure security is turned off when security is inserted,
            # else we may loop for ever...
            if self.cnx.transaction_data.get("security-rqlst-cache"):
                key = self.cache_key
            else:
                key = None
            if key is not None and key in self.cnx.transaction_data:
                cachedunion, args = self.cnx.transaction_data[key]
                union.children[:] = []
                for select in cachedunion.children:
                    union.append(select)
                union.has_text_query = cachedunion.has_text_query
                args.update(self.args)
                self.args = args
                cached = True
            else:
                with self.cnx.security_enabled(read=False):
                    noinvariant = self._insert_security(union)
                if key is not None:
                    self.cnx.transaction_data[key] = (union, self.args)
        else:
            noinvariant = ()
        if cached is None:
            self.cnx.vreg.rqlhelper.simplify(union)
            RQLAnnotator(self.schema).annotate(union)
            set_qdata(self.schema.relation_schema_for, union, noinvariant)
        if union.has_text_query:
            self.cache_key = None

    def _insert_security(self, union):
        noinvariant = set()
        for select in union.children[:]:
            for subquery in select.with_:
                self._insert_security(subquery.query)
            localchecks, restricted = self._check_permissions(select)
            if any(localchecks):
                self.cnx.rql_rewriter.insert_local_checks(
                    select, self.args, localchecks, restricted, noinvariant
                )
        return noinvariant

    def _check_permissions(self, rqlst):
        """Return a dict defining "local checks", i.e. RQLExpression defined in
        the schema that should be inserted in the original query, together with
        a set of variable names which requires some security to be inserted.

        Solutions where a variable has a type which the user can't definitly
        read are removed, else if the user *may* read it (i.e. if an rql
        expression is defined for the "read" permission of the related type),
        the local checks dict is updated.

        The local checks dict has entries for each different local check
        necessary, with associated solutions as value, a local check being
        defined by a list of 2-uple (variable name, rql expressions) for each
        variable which has to be checked. Solutions which don't require local
        checks will be associated to the empty tuple key.

        Note rqlst should not have been simplified at this point.
        """
        cnx = self.cnx
        msgs = []
        # dict(varname: eid), allowing to check rql expression for variables
        # which have a known eid
        varkwargs = {}
        if not cnx.transaction_data.get("security-rqlst-cache"):
            for var in rqlst.defined_vars.values():
                if var.stinfo["constnode"] is not None:
                    eid = var.stinfo["constnode"].eval(self.args)
                    varkwargs[var.name] = int(eid)
        # dictionary of variables restricted for security reason
        localchecks = {}
        restricted_vars = set()
        newsolutions = []
        for solution in rqlst.solutions:
            try:
                localcheck = get_local_checks(cnx, rqlst, solution)
            except Unauthorized as ex:
                msg = "remove %s from solutions since %s has no %s access to %s"
                msg %= (solution, cnx.user.login, ex.args[0], ex.args[1])
                msgs.append(msg)
                LOGGER.info(msg)
            else:
                newsolutions.append(solution)
                # try to benefit of rqlexpr.check cache for entities which
                # are specified by eid in query'args
                for varname, eid in varkwargs.items():
                    try:
                        rqlexprs = localcheck.pop(varname)
                    except KeyError:
                        continue
                    # if entity has been added in the current transaction, the
                    # user can read it whatever rql expressions are associated
                    # to its type
                    if cnx.added_in_transaction(eid):
                        continue
                    for rqlexpr in rqlexprs:
                        if rqlexpr.check(cnx, eid):
                            break
                    else:
                        raise Unauthorized(
                            "No read access on %r with eid %i." % (var, eid)
                        )
                # mark variables protected by an rql expression
                restricted_vars.update(localcheck)
                # turn local check into a dict key
                localcheck = tuple(sorted(localcheck.items()))
                localchecks.setdefault(localcheck, []).append(solution)
        # raise Unautorized exception if the user can't access to any solution
        if not newsolutions:
            raise Unauthorized("\n".join(msgs))
        # if there is some message, solutions have been modified and must be
        # reconsidered by the syntax treee
        if msgs:
            rqlst.set_possible_types(newsolutions)
        return localchecks, restricted_vars

    def finalize(self, select, solutions, insertedvars):
        rqlst = Union()
        rqlst.append(select)
        for mainvarname, rschema, newvarname in insertedvars:
            nvartype = str(rschema.objects(solutions[0][mainvarname])[0])
            for sol in solutions:
                sol[newvarname] = nvartype
        select.clean_solutions(solutions)
        add_types_restriction(self.schema, select)
        self.cnx.vreg.rqlhelper.annotate(rqlst)
        self.preprocess(rqlst, security=False)
        return rqlst


class InsertPlan(ExecutionPlan):
    """an execution model specific to the INSERT rql query"""

    def __init__(self, schema, rqlst, args, cnx):
        ExecutionPlan.__init__(self, schema, rqlst, args, cnx)
        # save originally selected variable, we may modify this
        # dictionary for substitution (query parameters)
        self.selected = rqlst.selection
        # list of rows of entities definition (ssplanner.EditedEntity)
        self.e_defs = [[]]
        # list of new relation definition (3-uple (from_eid, r_type, to_eid)
        self.r_defs = set()
        # indexes to track entity definitions bound to relation definitions
        self._r_subj_index = {}
        self._r_obj_index = {}
        self._expanded_r_defs = {}

    def add_entity_def(self, edef):
        """add an entity definition to build"""
        self.e_defs[-1].append(edef)

    def add_relation_def(self, rdef):
        """add an relation definition to build"""
        edef, rtype, value = rdef
        if self.schema[rtype].rule:
            raise QueryError(f"'{rtype}' is a computed relation")
        self.r_defs.add(rdef)
        if not isinstance(edef, int):
            self._r_subj_index.setdefault(edef, []).append(rdef)
        if not isinstance(value, int):
            self._r_obj_index.setdefault(value, []).append(rdef)

    def substitute_entity_def(self, edef, edefs):
        """substitute an incomplete entity definition by a list of complete
        equivalents

        e.g. on queries such as ::
          INSERT Personne X, Societe Y: X nom N, Y nom 'toto', X travaille Y
          WHERE U login 'admin', U login N

        X will be inserted as many times as U exists, and so the X travaille Y
        relations as to be added as many time as X is inserted
        """
        if not edefs or not self.e_defs:
            # no result, no entity will be created
            self.e_defs = ()
            return
        # first remove the incomplete entity definition
        colidx = self.e_defs[0].index(edef)
        for i, row in enumerate(self.e_defs[:]):
            self.e_defs[i][colidx] = edefs[0]
            samplerow = self.e_defs[i]
            for edef_ in edefs[1:]:
                row = [ed.clone() for i, ed in enumerate(samplerow) if i != colidx]
                row.insert(colidx, edef_)
                self.e_defs.append(row)
        # now, see if this entity def is referenced as subject in some relation
        # definition
        if edef in self._r_subj_index:
            for rdef in self._r_subj_index[edef]:
                expanded = self._expanded(rdef)
                result = []
                for exp_rdef in expanded:
                    for edef_ in edefs:
                        result.append((edef_, exp_rdef[1], exp_rdef[2]))
                self._expanded_r_defs[rdef] = result
        # and finally, see if this entity def is referenced as object in some
        # relation definition
        if edef in self._r_obj_index:
            for rdef in self._r_obj_index[edef]:
                expanded = self._expanded(rdef)
                result = []
                for exp_rdef in expanded:
                    for edef_ in edefs:
                        result.append((exp_rdef[0], exp_rdef[1], edef_))
                self._expanded_r_defs[rdef] = result

    def _expanded(self, rdef):
        """return expanded value for the given relation definition"""
        try:
            return self._expanded_r_defs[rdef]
        except KeyError:
            self.r_defs.remove(rdef)
            return [rdef]

    def relation_defs(self):
        """return the list for relation definitions to insert"""
        for rdefs in self._expanded_r_defs.values():
            yield from rdefs
        yield from self.r_defs

    def insert_entity_defs(self):
        """return eids of inserted entities in a suitable form for the resulting
        result set, e.g.:

        e.g. on queries such as ::
          INSERT Personne X, Societe Y: X nom N, Y nom 'toto', X travaille Y
          WHERE U login 'admin', U login N

        if there is two entities matching U, the result set will look like
        [(eidX1, eidY1), (eidX2, eidY2)]
        """
        cnx = self.cnx
        repo = cnx.repo
        results = []
        for row in self.e_defs:
            results.append([repo.glob_add_entity(cnx, edef) for edef in row])
        return results

    def insert_relation_defs(self):
        cnx = self.cnx
        repo = cnx.repo
        edited_entities = {}
        relations = {}
        for subj, rtype, obj in self.relation_defs():
            # if a string is given into args instead of an int, we get it here
            if isinstance(subj, str):
                subj = int(subj)
            elif not isinstance(subj, int):
                subj = subj.entity.eid
            if isinstance(obj, str):
                obj = int(obj)
            elif not isinstance(obj, int):
                obj = obj.entity.eid
            if repo.schema.relation_schema_for(rtype).inlined:
                if subj not in edited_entities:
                    entity = cnx.entity_from_eid(subj)
                    edited = EditedEntity(entity)
                    edited_entities[subj] = edited
                else:
                    edited = edited_entities[subj]
                edited.edited_attribute(rtype, obj)
            else:
                if rtype in relations:
                    relations[rtype].append((subj, obj))
                else:
                    relations[rtype] = [(subj, obj)]
        repo.glob_add_relations(cnx, relations)
        for edited in edited_entities.values():
            repo.glob_update_entity(cnx, edited)


class QuerierHelper:
    """helper class to execute rql queries, putting all things together"""

    def __init__(self, repo, schema):
        # system info helper
        self._repo = repo
        # instance schema
        self.set_schema(schema)

    def set_schema(self, schema):
        self.schema = schema
        self.clear_caches()

    def clear_caches(self, eids=None, etypes=None):
        if eids is None:
            self.rql_cache = RQLCache(self._repo, self.schema)
        else:
            cache = self.rql_cache
            for eid, etype in zip(eids, etypes):
                cache.pop((f"Any X WHERE X eid {eid}",), None)
                if etype is not None:
                    cache.pop((f"{etype} X WHERE X eid {eid}",), None)

    def plan_factory(self, rqlst, args, cnx):
        """create an execution plan for an INSERT RQL query"""
        if rqlst.TYPE == "insert":
            return InsertPlan(self.schema, rqlst, args, cnx)
        return ExecutionPlan(self.schema, rqlst, args, cnx)

    def _apply_function_on_results(self, cnx, results, selected, solution, rql):
        if len(results) == 0:
            return results
        functions_to_apply = [
            (i, varref)
            for (i, varref) in enumerate(selected)
            if isinstance(varref, (VariableRefMethodCall, VariableRefAttributeAccess))
        ]
        if len(functions_to_apply) == 0:
            return results
        new_results = deepcopy(results)
        for result in new_results:
            for i, varref in functions_to_apply:
                eid = result[i]
                entity = cnx.entity_from_eid(eid).cw_adapt_to("IRQLInterface")
                if isinstance(varref, VariableRefMethodCall):
                    try:
                        value = getattr(entity, varref.method_to_call)(
                            *[arg.value for arg in varref.args_for_method_to_call]
                        )
                    except AttributeError:
                        raise BadRQLQuery(
                            f"Function {varref.method_to_call} does not exists for "
                            f"EntityType {solution.get(varref.variable.name)} \n"
                            f"    RQL: {rql}"
                        )
                elif isinstance(varref, VariableRefAttributeAccess):
                    try:
                        value = getattr(entity, varref.attribute_to_access)
                    except AttributeError:
                        raise BadRQLQuery(
                            f"Attribute {varref.attribute_to_access} does not exists for "
                            f"EntityType {solution.get(varref.variable.name)} \n"
                            f"    RQL: {rql}"
                        )
                assert value  # test the value is well defined
                result[i] = value  # change with the function applied value
        return new_results

    def execute(self, cnx, rql, args=None, build_descr=True):
        """execute a rql query, return resulting rows and their description in
        a `ResultSet` object

        * `rql` should be a Unicode string or a plain ASCII string
        * `args` the optional parameters dictionary associated to the query
        * `build_descr` is a boolean flag indicating if the description should
          be built on select queries (if false, the description will be en empty
          list)

        on INSERT queries, there will be one row with the eid of each inserted
        entity

        result for DELETE and SET queries is undefined yet

        to maximize the rql parsing/analyzing cache performance, you should
        always use substitute arguments in queries (i.e. avoid query such as
        'Any X WHERE X eid 123'!)
        """
        use_debug_channel = debug_channel_has_subscribers("rql")

        if server.DEBUG & (server.DBG_RQL | server.DBG_SQL):
            if server.DEBUG & (server.DBG_MORE | server.DBG_SQL):
                print("*" * 80)
            print(
                "querier input", highlight_terminal(repr(rql)[1:-1], "RQL"), repr(args)
            )
        try:
            rqlst, cachekey = self.rql_cache.get(cnx, rql, args)
        except UnknownEid:
            # we want queries such as "Any X WHERE X eid 9999"
            # return an empty result instead of raising UnknownEid
            return empty_rset(rql, args)
        if rqlst.TYPE != "select":
            if cnx.read_security:
                check_no_password_selected(rqlst)
            cachekey = None
        else:
            if cnx.read_security:
                for select in rqlst.children:
                    check_no_password_selected(select)
                    check_relations_read_access(cnx, select, args)
            # on select query, always copy the cached rqlst so we don't have to
            # bother modifying it. This is not necessary on write queries since
            # a new syntax tree is built from them.
            rqlst = rqlst.copy()
            # Rewrite computed relations
            rewriter = RQLRelationRewriter(cnx)
            rewriter.rewrite(rqlst, args)
            self._repo.vreg.rqlhelper.annotate(rqlst)
            if args:
                # different SQL generated when some argument is None or not (IS
                # NULL). This should be considered when computing sql cache key
                cachekey += tuple(sorted([k for k, v in args.items() if v is None]))
        # make an execution plan
        plan = self.plan_factory(rqlst, args, cnx)
        plan.cache_key = cachekey
        plan.rql_query_tracing_token = str(uuid.uuid4())
        prepare_plan(plan, self.schema, self._repo.vreg.rqlhelper)

        if use_debug_channel:
            query_debug_informations = {
                "rql": rql,
                "rql_query_tracing_token": plan.rql_query_tracing_token,
                "args": args,
                # remove the last part of the stack which is: this line
                "callstack": "".join(traceback.format_stack()[:-1]),
                "description": "",
            }

        start = time.time()
        # execute the plan
        try:
            results = plan.execute()
        except (Unauthorized, ValidationError):
            # getting an Unauthorized/ValidationError exception means the
            # transaction must be rolled back
            #
            # notes:
            # * we should not reset the connections set here, since we don't want the
            #   connection to loose it during processing
            # * don't rollback if we're in the commit process, will be handled
            #   by the connection
            if cnx.commit_state is None:
                cnx.commit_state = "uncommitable"
            raise

        if use_debug_channel:
            query_debug_informations["time"] = (time.time() - start) * 1000
            query_debug_informations["result"] = results

        # build a description for the results if necessary
        descr = ()
        variables = None
        if build_descr:
            if rqlst.TYPE == "select":
                # sample selection
                if len(rqlst.children) == 1 and len(rqlst.children[0].solutions) == 1:
                    # easy, all lines are identical
                    selected = rqlst.children[0].selection
                    solution = rqlst.children[0].solutions[0]
                    description = _make_description(selected, args, solution)
                    descr = RepeatList(len(results), tuple(description))
                    variables = [
                        self._get_projected_name(projected, rqlst.children[0].stinfo)
                        for projected in selected
                    ]
                    results = self._apply_function_on_results(
                        cnx, results, selected, solution, rql
                    )

                else:
                    # hard, delegate the work :o)
                    descr = manual_build_descr(cnx, rqlst, args, results)
            elif rqlst.TYPE == "insert":
                # on insert plan, some entities may have been auto-casted,
                # so compute description manually even if there is only
                # one solution
                basedescr = [None] * len(plan.selected)
                todetermine = list(zip(range(len(plan.selected)), repeat(False)))
                descr = _build_descr(cnx, results, basedescr, todetermine)
            # FIXME: get number of affected entities / relations on non
            # selection queries ?
            if use_debug_channel:
                query_debug_informations["description"] = descr

        if use_debug_channel:
            emit_to_debug_channel("rql", query_debug_informations)

        # return a result set object
        return ResultSet(results, rql, args, descr, variables)

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None

    @staticmethod
    def _get_projected_name(projected, stinfo):
        if isinstance(projected, VariableRef):
            return projected.name
        elif isinstance(projected, Constant):
            if stinfo["rewritten"] is None:
                return str(projected)
            for name, value in stinfo["rewritten"].items():
                if [projected] == value:
                    return name
        return str(projected)


class RQLCache:
    def __init__(self, repo, schema):
        # rql st and solution cache.
        self._cache = LFUCache(repo.config["rql-cache-size"])
        self._lock = Lock()
        # rql cache key cache. Don't bother using a Cache instance: we should
        # have a limited number of queries in there, since there are no entries
        # in this cache for user queries (which have no args)
        self._ck_cache = {}
        # some cache usage stats
        self.cache_hit, self.cache_miss = 0, 0
        # rql parsing / analysing helper
        self.compute_var_types = repo.vreg.compute_var_types
        rqlhelper = repo.vreg.rqlhelper
        # set backend on the rql helper, will be used for function checking
        rqlhelper.backend = repo.config.system_source_config["db-driver"]

        def parse(rql, annotate=False, parse=rqlhelper.parse):
            """Return a freshly parsed syntax tree for the given RQL."""
            try:
                return parse(rql, annotate=annotate)
            except UnicodeError:
                raise RQLSyntaxError(rql)

        self._parse = parse

    def __len__(self):
        with self._lock:
            return len(self._cache)

    def get(self, cnx, rql, args):
        """Return syntax tree and cache key for the given RQL.

        Returned syntax tree is cached and must not be modified
        """
        with self._lock:
            # parse the query and binds variables
            cachekey = (rql,)
            try:
                if args:
                    # search for named args in query which are eids (hence
                    # influencing query's solutions)
                    eidkeys = self._ck_cache[rql]
                    if eidkeys:
                        # if there are some, we need a better cache key, eg (rql +
                        # entity type of each eid)
                        cachekey = _rql_cache_key(cnx, rql, args, eidkeys)
                rqlst = self._cache[cachekey]
                self.cache_hit += 1
            except KeyError:
                self.cache_miss += 1
                rqlst = self._parse(rql)
                # compute solutions for rqlst and return named args in query
                # which are eids. Notice that if you may not need `eidkeys`, we
                # have to compute solutions anyway (kept as annotation on the
                # tree)
                eidkeys = self.compute_var_types(cnx, rqlst, args)
                if args and rql not in self._ck_cache:
                    self._ck_cache[rql] = eidkeys
                    if eidkeys:
                        cachekey = _rql_cache_key(cnx, rql, args, eidkeys)
                self._cache[cachekey] = rqlst
            return rqlst, cachekey

    def pop(self, key, *args):
        """Pop a key from the cache."""
        with self._lock:
            self._cache.pop(key, *args)


def _rql_cache_key(cnx, rql, args, eidkeys):
    cachekey = [rql]
    type_from_eid = cnx.repo.type_from_eid
    for key in sorted(eidkeys):
        try:
            etype = type_from_eid(args[key], cnx)
        except KeyError:
            raise QueryError(f"bad cache key {key} (no value)")
        except TypeError:
            raise QueryError(f"bad cache key {key} (value: {args[key]!r})")
        cachekey.append(etype)
        # ensure eid is correctly typed in args
        args[key] = int(args[key])
    return tuple(cachekey)


LOGGER = getLogger("cubicweb.querier")
set_log_methods(QuerierHelper, LOGGER)


def manual_build_descr(cnx, rqlst, args, result):
    """build a description for a given result by analysing each row

    XXX could probably be done more efficiently during execution of query
    """
    # not so easy, looks for variable which changes from one solution
    # to another
    unstables = rqlst.get_variable_indices()
    basedescr = []
    todetermine = []
    for i in range(len(rqlst.children[0].selection)):
        ttype = _selection_idx_type(i, rqlst, args)
        if ttype is None or ttype == "Any":
            ttype = None
            isfinal = True
        else:
            isfinal = ttype in BASE_TYPES
        if ttype is None or i in unstables:
            basedescr.append(None)
            todetermine.append((i, isfinal))
        else:
            basedescr.append(ttype)
    if not todetermine:
        return RepeatList(len(result), tuple(basedescr))
    return _build_descr(cnx, result, basedescr, todetermine)


def _build_descr(cnx, result, basedescription, todetermine):
    description = []
    entity_type = cnx.entity_type
    todel = []
    for i, row in enumerate(result):
        row_descr = basedescription[:]
        for index, isfinal in todetermine:
            value = row[index]
            if value is None:
                # None value inserted by an outer join, no type
                row_descr[index] = None
                continue
            if isfinal:
                row_descr[index] = etype_from_pyobj(value)
            else:
                try:
                    row_descr[index] = entity_type(value)
                except UnknownEid:
                    cnx.error(
                        "wrong eid %s in repository, you should "
                        "db-check the database" % value
                    )
                    todel.append(i)
                    break
        else:
            description.append(tuple(row_descr))
    for i in reversed(todel):
        del result[i]
    return description


def _make_description(selected, args, solution):
    """return a description for a result set"""
    description = []
    for term in selected:
        description.append(term.get_type(solution, args))
    return description


def _selection_idx_type(i, rqlst, args):
    """try to return type of term at index `i` of the rqlst's selection"""
    for select in rqlst.children:
        term = select.selection[i]
        for solution in select.solutions:
            try:
                ttype = term.get_type(solution, args)
                if ttype is not None:
                    return ttype
            except CoercionError:
                return None



# copyright 2003-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
import os


def check_unused_index(cnx, min_usage=0):
    source = cnx.repo.system_source
    if source.dbdriver != "postgres":
        print("This command is only available for postgres.")
        return os.EX_UNAVAILABLE
    cursor = cnx.cnxset.cu
    cursor.execute(
        """
    SELECT
        s.relname AS tablename,
        s.indexrelname AS indexname,
        s.idx_scan AS indexusage,
        pg_size_pretty(pg_relation_size(s.indexrelid)) AS index_size,
        string_agg(a.attname, '|') as column_names
    FROM
        pg_catalog.pg_stat_all_indexes s
        JOIN pg_catalog.pg_index i ON s.indexrelid = i.indexrelid
        JOIN pg_attribute a ON (a.attnum = ANY(i.indkey) and i.indrelid = a.attrelid)
    WHERE
        s.idx_scan <= %(min_usage)s
        AND NOT EXISTS          -- does not enforce a constraint
        (
            SELECT 1 FROM pg_catalog.pg_constraint c
            WHERE c.conindid = s.indexrelid
        )
    GROUP BY s.relname, s.indexrelname, s.indexrelid, s.idx_scan
    ORDER BY pg_relation_size(s.indexrelid) DESC;
    """,
        {"min_usage": min_usage},
    )
    results = {
        "custom_indexes": [],
        "schema_indexes": [],
    }
    for tablename, index_name, index_size, column_names in cursor.fetchall():
        if not column_names:
            continue
        if index_name.startswith("pg_"):
            continue
        columns = column_names.split("|")
        if (
            not tablename.startswith("cw_")
            or len(columns) > 1
            or not column_names.startswith("cw_")
        ):
            # not an etype table
            results["custom_indexes"].append((tablename, index_name, index_size))
            continue
        etype = tablename[3:]
        attribute = column_names[3:]
        results["schema_indexes"].append((etype, attribute, index_size))
    if not (results["schema_indexes"] or results["schema_indexes"]):
        print("All indexes are used :)")
        return os.EX_OK
    if results["schema_indexes"]:
        print("Unused indexation in the schema:")
        for etype, attribute, size in results["schema_indexes"]:
            print(f"entity {etype.capitalize()}: attribute {attribute} (use {size})")
    if results["custom_indexes"]:
        print("Unused custom indexes:")
        for table, index_name, size in results["custom_indexes"]:
            print(f"table: {table}, index name: {index_name} (use {size})")
    return os.EX_OK



from cubicweb.server.sources import datafeed
from cubicweb.entities import AnyEntity
from cubicweb.entity import EntityAdapter
from cubicweb.predicates import is_instance


class SourceParserSuccess(datafeed.DataFeedParser):
    __regid__ = "test_source_parser_success"

    def process(self, url, raise_on_error=False):
        entity = self._cw.create_entity("Card", title="success")
        self.notify_updated(entity)


class SourceParserFail(SourceParserSuccess):
    __regid__ = "test_source_parser_fail"

    def process(self, url, raise_on_error=False):
        entity = self._cw.create_entity("Card", title="fail")
        self.notify_updated(entity)
        raise RuntimeError("fail")


class Affaire(AnyEntity):
    __regid__ = "Affaire"

    def ref_sujet(self):
        return f"{self.ref} -- {self.sujet}"


class AffaireRQLInterfaceAdapter(EntityAdapter):
    __regid__ = "IRQLInterface"
    __select__ = is_instance("Affaire")

    @property
    def sujet(self):
        return self.entity.sujet

    def ref_sujet_interface(self):
        return self.entity.ref_sujet()



import logging
from xml.etree import ElementTree

import markdown

log = logging.getLogger(__name__)

# Global Vars
URLIZE_RE = "(%s)" % "|".join(
    [
        r"<(?:f|ht)tps?://[^>]*>",
        r"\b(?:f|ht)tps?://[^)<>\s]+[^.,)<>\s]",
        r"\bwww\.[^)<>\s]+[^.,)<>\s]",
        r"[^(<\s]+\.(?:com|net|org)\b",
    ]
)


class UrlizePattern(markdown.inlinepatterns.Pattern):
    """This code comes from https://github.com/r0wb0t/markdown-urlize,
    and is made compatible with Python Markdown 3.4."""

    def handleMatch(self, m):
        url = m.group(2)

        if url.startswith("<"):
            url = url[1:-1]

        text = url

        if url.split("://")[0] not in ("http", "https", "ftp"):
            if "@" in url and "/" not in url:
                url = "mailto:" + url
            else:
                url = "http://" + url

        el = ElementTree.Element("a")
        el.set("href", url)
        el.text = markdown.util.AtomicString(text)
        return el


class UrlizeExtension(markdown.Extension):
    """Urlize Extension for Python-Markdown."""

    def extendMarkdown(self, md):
        """Replace autolink with UrlizePattern"""
        md.inlinePatterns.register(UrlizePattern(URLIZE_RE, md), "autolink", 120)


def markdown_publish(context, data):
    """publish a string formatted as MarkDown Text to HTML

    :type context: a cubicweb application object

    :type data: str
    :param data: some MarkDown text

    :rtype: unicode
    :return:
      the data formatted as HTML or the original data if an error occurred
    """
    md = markdown.Markdown(extensions=["extra", UrlizeExtension()])
    try:
        return md.convert(data)
    except Exception:
        import traceback

        traceback.print_exc()
        log.exception("Error while converting Markdown to HTML")
        return data



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""custom storages for the system source"""

import os
import tempfile
from contextlib import contextmanager
from os import unlink, path as osp

from logilab.common import nullobject
from yams.schema import role_name

from yams import ValidationError
from cubicweb import Binary
from cubicweb.server import hook
from cubicweb.server.edition import EditedEntity


def set_attribute_storage(repo, etype, attr, storage):
    repo.system_source.set_storage(etype, attr, storage)


def unset_attribute_storage(repo, etype, attr):
    repo.system_source.unset_storage(etype, attr)


class Storage:
    """abstract storage

    * If `source_callback` is true (by default), the callback will be run during
      query result process of fetched attribute's value and should have the
      following prototype::

        callback(self, source, cnx, value)

      where `value` is the value actually stored in the backend. None values
      will be skipped (eg callback won't be called).

    * if `source_callback` is false, the callback will be run during sql
      generation when some attribute with a custom storage is accessed and
      should have the following prototype::

        callback(self, generator, relation, linkedvar)

      where `generator` is the sql generator, `relation` the current rql syntax
      tree relation and linkedvar the principal syntax tree variable holding the
      attribute.
    """

    is_source_callback = True

    def callback(self, *args):
        """see docstring for prototype, which vary according to is_source_callback"""
        raise NotImplementedError()

    def entity_added(self, entity, attr):
        """an entity using this storage for attr has been added"""
        raise NotImplementedError()

    def entity_updated(self, entity, attr):
        """an entity using this storage for attr has been updatded"""
        raise NotImplementedError()

    def entity_deleted(self, entity, attr):
        """an entity using this storage for attr has been deleted"""
        raise NotImplementedError()

    def migrate_entity(self, entity, attribute):
        """migrate an entity attribute to the storage"""
        raise NotImplementedError()


# TODO
# * make it configurable without code
# * better file path attribution
# * handle backup/restore


def uniquify_path(dirpath, basename):
    """return a file descriptor and unique file name for `basename` in `dirpath`"""
    path = basename.replace(osp.sep, "-")
    base, ext = osp.splitext(path)
    return tempfile.mkstemp(prefix=base, suffix=ext, dir=dirpath)


@contextmanager
def fsimport(cnx):
    present = "fs_importing" in cnx.transaction_data
    old_value = cnx.transaction_data.get("fs_importing")
    cnx.transaction_data["fs_importing"] = True
    yield
    if present:
        cnx.transaction_data["fs_importing"] = old_value
    else:
        del cnx.transaction_data["fs_importing"]


_marker = nullobject()


class BytesFileSystemStorage(Storage):
    """store Bytes attribute value on the file system"""

    def __init__(self, defaultdir, fsencoding=_marker, wmode=0o444):
        if fsencoding is not _marker:
            raise ValueError("fsencoding is no longer supported in python 3")
        self.default_directory = defaultdir
        # extra umask to use when creating file
        # 0444 as in "only allow read bit in permission"
        self._wmode = wmode

    def _writecontent(self, fd, binary):
        """write the content of a binary in readonly file

        As the bfss never alters an existing file it does not prevent it from
        working as intended. This is a better safe than sorry approach.
        """
        os.fchmod(fd, self._wmode)
        fileobj = os.fdopen(fd, "wb")
        binary.to_file(fileobj)
        fileobj.close()

    def callback(self, source, cnx, value):
        """sql generator callback when some attribute with a custom storage is
        accessed
        """
        fpath = source.binary_to_str(value)
        try:
            return Binary.from_file(fpath)
        except OSError as ex:
            source.critical("can't open %s: %s", value, ex)
            return None

    def entity_added(self, entity, attr):
        """an entity using this storage for attr has been added"""
        if entity._cw.transaction_data.get("fs_importing"):
            binary = Binary.from_file(entity.cw_edited[attr].getvalue())
            entity._cw_dont_cache_attribute(attr, repo_side=True)
        else:
            binary = entity.cw_edited.pop(attr)
            if binary is not None:
                fd, fpath = self.new_fs_path(entity, attr)
                # bytes storage used to store file's path
                binary_obj = Binary(fpath.encode("utf-8"))
                entity.cw_edited.edited_attribute(attr, binary_obj)
                self._writecontent(fd, binary)
                AddFileOp.get_instance(entity._cw).add_data(fpath)
        return binary

    def entity_updated(self, entity, attr):
        """an entity using this storage for attr has been updated"""
        # get the name of the previous file containing the value
        oldpath = self.current_fs_path(entity, attr)
        if entity._cw.transaction_data.get("fs_importing"):
            # If we are importing from the filesystem, the file already exists.
            # We do not need to create it but we need to fetch the content of
            # the file as the actual content of the attribute
            fpath = entity.cw_edited[attr].getvalue()
            entity._cw_dont_cache_attribute(attr, repo_side=True)
            assert fpath is not None
            binary = Binary.from_file(fpath)
        else:
            # We must store the content of the attributes
            # into a file to stay consistent with the behaviour of entity_add.
            # Moreover, the BytesFileSystemStorage expects to be able to
            # retrieve the current value of the attribute at anytime by reading
            # the file on disk. To be able to rollback things, use a new file
            # and keep the old one that will be removed on commit if everything
            # went ok.
            #
            # fetch the current attribute value in memory
            binary = entity.cw_edited.pop(attr)
            if binary is None:
                fpath = None
            else:
                # Get filename for it
                fd, fpath = self.new_fs_path(entity, attr)
                # write attribute value on disk
                self._writecontent(fd, binary)
                # Mark the new file as added during the transaction.
                # The file will be removed on rollback
                AddFileOp.get_instance(entity._cw).add_data(fpath)
            # reinstall poped value
            if fpath is None:
                entity.cw_edited.edited_attribute(attr, None)
            else:
                # register the new location for the file.
                binary_obj = Binary(fpath.encode("utf-8"))
                entity.cw_edited.edited_attribute(attr, binary_obj)
        if oldpath is not None and oldpath != fpath:
            # Mark the old file as useless so the file will be removed at
            # commit.
            DeleteFileOp.get_instance(entity._cw).add_data(oldpath)
        return binary

    def entity_deleted(self, entity, attr):
        """an entity using this storage for attr has been deleted"""
        fpath = self.current_fs_path(entity, attr)
        if fpath is not None:
            DeleteFileOp.get_instance(entity._cw).add_data(fpath)

    def new_fs_path(self, entity, attr):
        # We try to get some hint about how to name the file using attribute's
        # name metadata, so we use the real file name and extension when
        # available. Keeping the extension is useful for example in the case of
        # PIL processing that use filename extension to detect content-type, as
        # well as providing more understandable file names on the fs.
        basename = [str(entity.eid), attr]
        name = entity.cw_attr_metadata(attr, "name")
        if name is not None:
            basename.append(name)
        fd, fspath = uniquify_path(self.default_directory, "_".join(basename))
        if fspath is None:
            msg = entity._cw._("failed to uniquify path (%s, %s)") % (
                self.default_directory,
                "_".join(basename),
            )
            raise ValidationError(entity.eid, {role_name(attr, "subject"): msg})
        assert isinstance(fspath, str)
        return fd, fspath

    def current_fs_path(self, entity, attr):
        """return the current fs_path of the attribute, or None is the attr is
        not stored yet.
        """
        sysource = entity._cw.repo.system_source
        cu = sysource.doexec(
            entity._cw,
            f"SELECT cw_{attr} FROM cw_{entity.cw_etype} WHERE cw_eid={entity.eid}",
        )
        rawvalue = cu.fetchone()[0]
        if rawvalue is None:  # no previous value
            return None
        fspath = sysource._process_value(rawvalue, cu.description[0], binarywrap=bytes)
        return fspath.decode("utf-8")

    def migrate_entity(self, entity, attribute):
        """migrate an entity attribute to the storage"""
        entity.cw_edited = EditedEntity(entity, **entity.cw_attr_cache)
        binary = self.entity_added(entity, attribute)
        if binary is not None:
            cnx = entity._cw
            source = cnx.repo.system_source
            attrs = source.preprocess_entity(entity)
            sql = source.sqlgen.update("cw_" + entity.cw_etype, attrs, ["cw_eid"])
            source.doexec(cnx, sql, attrs)
        entity.cw_edited = None


class AddFileOp(hook.DataOperationMixIn, hook.Operation):
    def rollback_event(self):
        for filepath in self.get_data():
            assert isinstance(filepath, str)
            try:
                unlink(filepath)
            except Exception as ex:
                self.error(f"can't remove {filepath}: {ex}")


class DeleteFileOp(hook.DataOperationMixIn, hook.Operation):
    def postcommit_event(self):
        for filepath in self.get_data():
            assert isinstance(filepath, str)
            try:
                unlink(filepath)
            except Exception as ex:
                self.error(f"can't remove {filepath}: {ex}")



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.predicates import is_instance
from cubicweb.server import hook


class SecurityHooksTC(CubicWebTC):
    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            self.add_eid = cnx.create_entity(
                "EmailAddress", address="hop@perdu.com", reverse_use_email=cnx.user.eid
            ).eid
            cnx.commit()

    def test_inlined_cw_edited_relation(self):
        """modification of cw_edited to add an inlined relation shouldn't trigger a security error.

        Test for https://www.cubicweb.org/ticket/5477315
        """
        sender = self.repo.schema["Email"].relation_definition("sender")
        with self.temporary_permissions((sender, {"add": ()})):

            class MyHook(hook.Hook):
                __regid__ = "test.pouet"
                __select__ = hook.Hook.__select__ & is_instance("Email")
                events = ("before_add_entity",)

                def __call__(self):
                    self.entity.cw_edited["sender"] = self._cw.user.primary_email[0].eid

            with self.temporary_appobjects(MyHook):
                with self.admin_access.repo_cnx() as cnx:
                    email = cnx.create_entity("Email", messageid="1234")
                    cnx.commit()
                    self.assertEqual(email.sender[0].eid, self.add_eid)


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



from cubicweb.entities import AnyEntity, fetch_config


class BlogEntry(AnyEntity):
    __regid__ = "BlogEntry"
    fetch_attrs, cw_fetch_order = fetch_config(["creation_date", "title"], order="DESC")
    rest_attr = "unique_id"


class PublicBlogEntry(AnyEntity):
    __regid__ = "PublicBlogEntry"
    fetch_attrs, cw_fetch_order = fetch_config(["creation_date", "title"], order="DESC")
    rest_attr = "unique_id"



from base64 import b64decode


# before removing extid, ensure it's coherent with cwuri
for eid, etype, encoded_extid in sql(
    "SELECT eid, type, extid FROM entities, cw_CWSource "
    "WHERE cw_CWSource.cw_name=entities.asource AND cw_CWSource.cw_type='ldapfeed'"
):
    sql(
        f"UPDATE cw_{etype} SET cw_cwuri=%(cwuri)s WHERE cw_eid=%(eid)s",
        {"eid": eid, "cwuri": b64decode(encoded_extid)},
    )

sql("DROP TABLE IF EXISTS moved_entities")
sql("ALTER TABLE entities DROP COLUMN asource")
sql("ALTER TABLE entities DROP COLUMN extid")
sql("DROP INDEX entities_type_idx")

# force cw_schema deletion before CWSourceSchemaConfig to avoid nasty bug
drop_relation_type("cw_schema")
drop_entity_type("CWSourceSchemaConfig")

# Check the CW versions and add the entity only if needed ?
add_entity_type("CWSession")
rql('DELETE CWProperty X WHERE X pkey "system.version.pyramid"', ask_confirm=False)



# copyright 2019 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from rdflib import namespace


RDF_MIMETYPE_TO_FORMAT = {
    "application/rdf+xml": "xml",
    "text/turtle": "turtle",
    "text/n3": "n3",
    "application/n-quads": "nquads",
    "application/n-triples": "nt",
    "application/trig": "trig",
    "application/ld+json": "json-ld",
}

NAMESPACES = {
    "rdf": namespace.RDF,
    "rdfs": namespace.RDFS,
    "owl": namespace.OWL,
    "xsd": namespace.XSD,
    "skos": namespace.SKOS,
    "void": namespace.VOID,
    "dc": namespace.DC,
    "dcterms": namespace.DCTERMS,
    "foaf": namespace.FOAF,
    "doap": namespace.DOAP,
    "schema": namespace.Namespace("https://schema.org/"),
    "cubicweb": namespace.Namespace("http://ns.cubicweb.org/cubicweb/0.0/"),
}


def add_entity_to_graph(graph, entity, adapter_regid="rdf"):
    adapter = entity.cw_adapt_to(adapter_regid)
    if adapter:
        for triple in adapter.triples():
            graph.add(triple)
        for prefix, rdfns in adapter.used_namespaces.items():
            graph.bind(prefix, rdfns)



# copyright 2010-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""hooks for repository sources synchronization"""

from cubicweb import _

from cubicweb import validation_error
from cubicweb.predicates import is_instance
from cubicweb.server import SOURCE_TYPES, hook


class SourceHook(hook.Hook):
    __abstract__ = True
    category = "cw.sources"

    def get_source(self, source_entity):
        if source_entity.name == "system":
            return self._cw.repo.system_source
        return self._cw.repo.get_source(
            source_entity.type, source_entity.name, {}, source_entity.eid
        )


class SourceAddedHook(SourceHook):
    __regid__ = "cw.sources.added"
    __select__ = SourceHook.__select__ & is_instance("CWSource")
    events = ("after_add_entity",)

    def __call__(self):
        if self.entity.type not in SOURCE_TYPES:
            msg = _("Unknown source type")
            raise validation_error(self.entity, {("type", "subject"): msg})

        source = self.get_source(self.entity)
        source.check_urls(self.entity)
        source.check_config(self.entity)


class SourceRemovedHook(SourceHook):
    __regid__ = "cw.sources.removed"
    __select__ = SourceHook.__select__ & is_instance("CWSource")
    events = ("before_delete_entity",)

    def __call__(self):
        if self.entity.name == "system":
            msg = _("You cannot remove the system source")
            raise validation_error(self.entity, {None: msg})


class SourceUpdatedHook(SourceHook):
    __regid__ = "cw.sources.configupdate"
    __select__ = SourceHook.__select__ & is_instance("CWSource")
    events = ("before_update_entity",)

    def __call__(self):
        if "name" in self.entity.cw_edited:
            oldname, newname = self.entity.cw_edited.oldnewvalue("name")
            if oldname == "system":
                msg = _("You cannot rename the system source")
                raise validation_error(self.entity, {("name", "subject"): msg})

        source = self.get_source(self.entity)
        if "url" in self.entity.cw_edited:
            source.check_urls(self.entity)
        if "config" in self.entity.cw_edited:
            source.check_config(self.entity)
